import copy

import pandas as pnd

from ..commons.medium import apply_medium_given_column

from ..mkmodel.gapfillutils import can_synth



def biosynthesis_on_media(logger, model, dbexp, media, biosynth):
    
    
    # check if requested
    df_S = pnd.DataFrame()
    if type(biosynth) == bool and biosynth == False:  # mkmodel
        return df_S
    if type(biosynth) == str and biosynth == '-':  # parsedb
        return df_S
    
    
    # define metabolites to check:
    to_check = []
    if type(biosynth) == bool and biosynth == True:  # mkmodel
        for m in model.metabolites:
            if m.id.endswith('_c'):
                to_check.append(m)
    if type(biosynth) == str and biosynth != '-':
        provided_puremids = biosynth.split(',')
        available_puremids = [m.id.rsplit('_',1)[0] for m in model.metabolites if m.id.endswith('_c')]
        for puremid in copy.deepcopy(provided_puremids):
            if puremid not in available_puremids:
                logger.warning(f"Skipping metabolite '{puremid}' provided with --biosynth as it does not exist.")
                provided_puremids.remove(puremid)
        for puremid in provided_puremids:
            to_check.append(model.metabolites.get_by_id(f"{puremid}_c"))
    if len(to_check) == 0:
        return df_S
    
    
    # prepare sheet for excel output
    df_S['name'] = None
    for m in model.metabolites:
        if m.id.endswith('_c'):
            df_S.loc[m.id, 'name'] = m.name
            
            
    # get involved media:
    if media == '-':
        logger.info(f"No media provided: metabolites biosynthesis check will be skipped.")
        return df_S
    media = media.split(',')
    # at least 1 medium must exist:
    if any([i in dbexp['media'].columns for i in media])==False:
        logger.error(f"None of the provided media IDs exist. Available media are {list(dbexp['media'].columns)}.")
        return 1
    
    
    logger.info(f"Checking biosynthesis of {len(to_check)} metabolites on {len(media)} media...")
    for medium in media:
        if medium not in dbexp['media'].columns:
            logger.info(f"Medium '{medium}' does not exists and will be ignored.")
            continue
        # create dedicated column on the excel output
        df_S[f'{medium}'] = None
        logger.debug(f"Checking metabolites biosynthesis on medium '{medium}'...")
            
            
        # apply medium :
        response = apply_medium_given_column(logger, model, medium, dbexp['media'][medium])
        if response == 1: return 1
            
    
        for m in to_check:
            dem = '/'
            binary, obj_value, status = can_synth(model, m.id)


            # check if it's dead-end metabolite (dem)
            if binary == False:
                dem = False
                is_consumed = False
                is_produced = False
                for r in m.reactions:
                    if m.id in [m2.id for m2 in r.reactants]:
                        is_consumed = True
                    if m.id in [m2.id for m2 in r.products]:
                        is_produced = True
                if   is_consumed and not is_produced:
                    dem = 'no_production'
                elif is_produced and not is_consumed:
                    dem = 'no_consumption' 


            if status == 'infeasible': 
                df_S.loc[m.id, f'{medium}'] = 'infeasible'
            elif obj_value == 0:
                if dem == '/':
                    df_S.loc[m.id, f'{medium}'] = 'blocked'
                else:
                    df_S.loc[m.id, f'{medium}'] = dem
            else:
                df_S.loc[m.id, f'{medium}'] = obj_value

    
    return df_S




import os
import warnings
import logging
import pickle
import logging
from importlib import resources

import pandas as pnd
import cobra

from ..commons.metrics import log_metrics
from ..commons.metrics import log_unbalances

from .gapfillutils import import_from_universe

from .polishing import remove_disconnected
from .polishing import remove_orphans
from .polishing import remove_sinks_demands

    
    
def load_input_universe(logger, universe):
    
    
    # built-in or fresh universe path
    path_to_unixml = None
    
    
    # use the buil-in one if no fresh universe wa provided
    if universe == '-':
        logger.info(f"No fresh universe provided: loading built-in whole universe...")
        with resources.path("gsrap.assets", f"universe.xml") as asset_path: 
            
            # this is the sbml to load
            path_to_unixml = asset_path
    
    
    # user provided a fresh universe
    else: 
        # check if files exist
        if os.path.isfile(universe) == False: 
            logger.error(f"Provided --universe doesn't exist: {universe}.")
            return 1
        
        # check extension
        if universe.endswith('.xml') == False:
            logger.error(f"Provided --universe must be in cobrapy-compatible SBML format (.xml extension).")
            return 1

        # this is the sbml to load
        logger.info(f"Loading provided fresh whole universe...")
        path_to_unixml = universe
        
        
    # Ignore eventual "https://identifiers.org/curator_notes/ does not conform to ..."
    with warnings.catch_warnings():  # temporarily suppress warnings for this block
        warnings.simplefilter("ignore")  # ignore all warnings

        # disable warnings
        cobra_logger = logging.getLogger("cobra.io.sbml")
        old_level = cobra_logger.level
        cobra_logger.setLevel(logging.ERROR)   

        # load the model
        universe = cobra.io.read_sbml_model(path_to_unixml)

        # restore original behaviour: 
        cobra_logger.setLevel(old_level)   
        
    
    # log main universe metrics:
    log_metrics(logger, universe)
    #log_unbalances(logger, universe)  # commented as it is 0 by design
        
        
    return universe



def prune_universe(logger, universe, taxon, transrule):
    
    
    # get available kingdoms and phyla
    avail_kingdoms = set()
    avail_phyla = set()
    for g in universe.genes:
        if 'kingdom' in g.annotation.keys():
            annot = g.annotation['kingdom']
            # saved as lists, resulting str after reading: now back to lists
            if type(annot) == str: annot = [annot]
            for i in annot:
                avail_kingdoms.add(i)
        if 'phylum' in g.annotation.keys():
            annot = g.annotation['phylum']
            # saved as lists, resulting str after reading: now back to lists
            if type(annot) == str: annot = [annot]
            for i in annot:
                avail_phyla.add(i)
                
    
    # check formatting
    if ':' not in taxon:
        logger.error(f"Format of provided --taxon is not compliant (missing ':').")
        return 1
    if len(taxon.split(':')) > 2:
        logger.error(f"Format of provided --taxon is not compliant (too many ':').")
        return 1
    level, name = taxon.split(':')
    allowed_levels = ['kingdom', 'phylum']
    if level not in allowed_levels:
        logger.error(f"Level '{level}' provided with --taxon is not allowed: please choose between {allowed_levels}.")
        return 1
    if level=='kingdom' and name not in avail_kingdoms:
        logger.error(f"Cannot find a taxon of level 'kingdom' named '{name}': please choose between {avail_kingdoms}.")
        return 1
    if level=='phylum' and name not in avail_phyla:
        logger.error(f"Cannot find a taxon of level 'phylum' named '{name}': please choose between {avail_phyla}.")
        return 1
        
    
    # create a copy of the universe
    logger.info(f"Pruning universe to represent taxon '{level}:{name}'...")
    pruned_universe = universe.copy()
    
    
    # remove sinks/demands
    n = remove_sinks_demands(logger, pruned_universe)   # none logger to avoid confusion.
        
        
    # mark orthologs for deletion
    to_remove = []
    for g in pruned_universe.genes:
        if level in g.annotation.keys():  # 'orphan' reactions remain!
            annot = g.annotation[level]
            # saved as lists, resulting str after reading: now back to lists
            if type(annot) == str: annot = [annot]
            if name not in annot:
                to_remove.append(g)
    logger.debug(f"Removing {len(to_remove)} non-taxon orthologs...")
           
        
    # delete marked orthologs
    # trick to avoid the WARNING "cobra/core/group.py:147: UserWarning: need to pass in a list" 
    # triggered when trying to remove reactions that are included in groups. 
    with warnings.catch_warnings():  # temporarily suppress warnings for this block
        warnings.simplefilter("ignore")  # ignore all warnings
        cobra_logger = logging.getLogger("cobra.util.solver")
        old_level = cobra_logger.level
        cobra_logger.setLevel(logging.ERROR)   

        cobra.manipulation.delete.remove_genes(pruned_universe, to_remove, remove_reactions=True)

        # restore original behaviour: 
        cobra_logger.setLevel(old_level) 
        
        
    # reintroduce transport reactions as orphans:
    if transrule == 'rescue': 
        rids_pruned_universe = [r.id for r in pruned_universe.reactions]
        for r in universe.reactions:
            if len(set([m.id.rsplit('_',1)[-1] for m in r.metabolites])) > 1:
                if r.id not in rids_pruned_universe:
                    logger.debug(f"Reintroducing trasport reaction '{r.id}'...")
                    import_from_universe(pruned_universe, universe, r.id, gpr='')  
                    
                    
    # remove universal orphans
    if transrule == 'rescue': keep_transport = True 
    elif transrule == 'ignore': keep_transport = False
    else: keep_transport = False
    n = remove_orphans(logger, pruned_universe, keep_transport)
                    
    
    # remove disconnected metabolites
    n = remove_disconnected(logger, pruned_universe)  # none logger to avoid confusion.
    
    
    # log main universe metrics:
    log_metrics(logger, pruned_universe)
    
    
    return pruned_universe



def load_input_eggnog(logger, eggnog):
    
    
    # load eggnog annotations
    df_eggnog = pnd.read_csv(eggnog, sep='\t', comment='#', header=None)
    df_eggnog.columns = 'query	seed_ortholog	evalue	score	eggNOG_OGs	max_annot_lvl	COG_category	Description	Preferred_name	GOs	EC	KEGG_ko	KEGG_Pathway	KEGG_Module	KEGG_Reaction	KEGG_rclass	BRITE	KEGG_TC	CAZy	BiGG_Reaction	PFAMs'.split('\t')
    df_eggnog = df_eggnog.set_index('query', drop=True, verify_integrity=True)
    
    
    return df_eggnog



def load_keggorg_like_eggnog(logger, keggorg, outdir):
    
    
    # load raw data, downloaded form kegg: 
    df_keggorg = pickle.load(open(os.path.join(outdir, f'{keggorg}.keggorg'), 'rb'))
    df_keggorg = df_keggorg.set_index('gid', drop=True, verify_integrity=True)


    # create an eggnog-like dataframe:
    df_eggnog_like = []   # list of dict future df
    for gid in df_keggorg.index:
        row_dict = {}

        row_dict['query'] = gid
        row_dict['PFAMs'] = ','.join(df_keggorg.loc[gid, 'Pfam']) if type(df_keggorg.loc[gid, 'Pfam'])==list else '-'
        row_dict['KEGG_ko'] = df_keggorg.loc[gid, 'ko'] if type(df_keggorg.loc[gid, 'ko'])==str else '-'

        df_eggnog_like.append(row_dict)
    df_eggnog_like = pnd.DataFrame.from_records(df_eggnog_like)


    # appen missing coluns and sort
    eggnog_columns = 'query	seed_ortholog	evalue	score	eggNOG_OGs	max_annot_lvl	COG_category	Description	Preferred_name	GOs	EC	KEGG_ko	KEGG_Pathway	KEGG_Module	KEGG_Reaction	KEGG_rclass	BRITE	KEGG_TC	CAZy	BiGG_Reaction	PFAMs'.split('\t')
    for c in eggnog_columns:  
        if c not in df_eggnog_like.columns:
            df_eggnog_like[c] = '-'
    df_eggnog_like = df_eggnog_like[eggnog_columns]


    # set the index like in eggnog
    df_eggnog_like = df_eggnog_like.set_index('query', drop=True, verify_integrity=True)
    return df_eggnog_like



def parse_eggnog(df_eggnog):
    
    
    # PART 1. get KO codes available
    gid_to_kos = {}
    ko_to_gids = {}
    for gid, kos in df_eggnog['KEGG_ko'].items():
        if kos == '-': 
            continue
            
        if gid not in gid_to_kos.keys(): 
            gid_to_kos[gid] = set()
            
        kos = kos.split(',')
        kos = [i.replace('ko:', '') for i in kos]
        for ko in kos: 
            if ko not in ko_to_gids.keys(): 
                ko_to_gids[ko] = set()
                
            # populate dictionaries
            ko_to_gids[ko].add(gid)
            gid_to_kos[gid].add(ko)

    
    return ko_to_gids, gid_to_kos



def get_modeled_kos(model):
    
    
    # get modeled KO ids:
    modeled_gid_to_ko = {}
    modeled_ko_to_gid = {}
    
    for g in model.genes:
        if g.id in ['orphan', 'spontaneous']: 
            continue
        corresponding_ko = g.annotation['ko']
        
        modeled_gid_to_ko[g.id] = corresponding_ko
        modeled_ko_to_gid[corresponding_ko] = g.id
        
    modeled_kos = list(modeled_gid_to_ko.values())
        
    return modeled_kos, modeled_gid_to_ko, modeled_ko_to_gid



def subtract_kos(logger, model, eggonog_ko_to_gids):
    
    
    modeled_kos, _, modeled_ko_to_gid = get_modeled_kos(model)
        
        
    to_remove = []  # genes to delete
    for ko in modeled_kos: 
        if ko not in eggonog_ko_to_gids.keys():
            gid_to_remove = modeled_ko_to_gid[ko]
            to_remove.append(model.genes.get_by_id(gid_to_remove))
            
    
    # remove also orphan reactions!
    to_remove.append(model.genes.get_by_id('orphan'))
    
    
    # delete marked genes!
    # trick to avoid the WARNING "cobra/core/group.py:147: UserWarning: need to pass in a list" 
    # triggered when trying to remove reactions that are included in groups. 
    with warnings.catch_warnings():  # temporarily suppress warnings for this block
        warnings.simplefilter("ignore")  # ignore all warnings
        cobra_logger = logging.getLogger("cobra.util.solver")
        old_level = cobra_logger.level
        cobra_logger.setLevel(logging.ERROR)   

        cobra.manipulation.delete.remove_genes(model, to_remove, remove_reactions=True)

        # restore original behaviour: 
        cobra_logger.setLevel(old_level)   
        
   
    logger.info(f"Found annotations for {len(model.genes)} orthologs modeled in universe.")
    return 0



def translate_remaining_kos(logger, model, eggonog_ko_to_gids):
    
    
    n_starting_orthologs = len(model.genes)
    _, modeled_gid_to_ko, _ = get_modeled_kos(model) 
    
    
    # iterate reactions:
    for r in model.reactions:

        gpr = r.gene_reaction_rule

        # force each gid to be surrounded by spaces: 
        gpr = ' ' + gpr.replace('(', ' ( ').replace(')', ' ) ') + ' '
        
        for gid in modeled_gid_to_ko.keys():
            if f' {gid} ' in gpr:
                
                new_gids = eggonog_ko_to_gids[modeled_gid_to_ko[gid]]
                gpr = gpr.replace(f' {gid} ', f' ({" or ".join(new_gids)}) ')       
            

        # remove spaces between parenthesis
        gpr = gpr.replace(' ( ', '(').replace(' ) ', ')')
        # remove spaces at the extremes: 
        gpr = gpr[1: -1]


        # New genes are introduced. Parethesis at the extremes are removed if not necessary. 
        r.gene_reaction_rule = gpr
        r.update_genes_from_gpr()
            
            
    # remaining old 'Cluster_'s need to removed.
    # remove if (1) hte ID starts with clusters AND (2) they are no more associated with any reaction
    to_remove = []
    for g in model.genes:
        
        if g.id in ['orphan', 'spontaneous']:
            continue
            
        if g.id in modeled_gid_to_ko.keys() and len(g.reactions)==0:
            to_remove.append(g)
            
    # warning suppression not needed here, as no reaction is actually removed.
    cobra.manipulation.delete.remove_genes(model, to_remove, remove_reactions=True)
    
        
    logger.info(f"Translated {n_starting_orthologs} orthologs to {len(model.genes)} strain-specific genes.")
    return 0
        
    

def restore_gene_annotations(logger, model, universe, eggonog_gid_to_kos):
    
    
    for g in model.genes:
        if g.id == 'spontaneous': 
            continue
            
        names = []    
        for ko in eggonog_gid_to_kos[g.id]:
            
            # get the corresponding universal gene:
            uni_g = None
            for ug in universe.genes:
                if 'ko' not in ug.annotation.keys():
                    continue
                if ug.annotation['ko']==ko:  # take the first (and only)
                    uni_g = ug
                    break
            if uni_g == None:  
                # The ko provided by eggnog-mapper is still not modeled in the universe.
                # Multiple ko are possible for each gene. Of these, only 1 could b modeled.
                continue
            
            
            # transfer annotations of this ko/universal gene:
            for key in uni_g.annotation.keys():
                if key == 'ko':
                    continue   # resulting models will loose links to kos.
                if key not in g.annotation:
                    g.annotation[key] = []
                items = uni_g.annotation[key]
                if type(items)==str:  items = [items]
                for i in items:
                    g.annotation[key].append(i)
                    
            # collect names
            names.append(uni_g.name)
        g.name = '; '.join(names)
        
        
        
def append_keggorg_gene_annots(logger, model, keggorg, outdir):
    

    # load raw data, downloaded form kegg: 
    df_keggorg = pickle.load(open(os.path.join(outdir, f'{keggorg}.keggorg'), 'rb'))
    df_keggorg = df_keggorg.set_index('gid', drop=True, verify_integrity=True)
    
    
    # KEGG can provide some useful (ie, used in Memote) gene annotations:
    for g in model.genes:
        if g.id in df_keggorg.index:
            
            g.annotation['kegg.genes'] = [keggorg + ':' + g.id]
            
            if 'NCBI-GeneID' in df_keggorg.columns:
                g.annotation['ncbigene'] = df_keggorg.loc[g.id, 'NCBI-GeneID'] if type(df_keggorg.loc[g.id, 'NCBI-GeneID'])==list else []
            if 'NCBI-ProteinID' in df_keggorg.columns:
                g.annotation['ncbiprotein'] = df_keggorg.loc[g.id, 'NCBI-ProteinID'] if type(df_keggorg.loc[g.id, 'NCBI-ProteinID'])==list else []
            if 'ASAP' in df_keggorg.columns:
                g.annotation['asap'] = df_keggorg.loc[g.id, 'ASAP'] if type(df_keggorg.loc[g.id, 'ASAP'])==list else []
            if 'UniProt' in df_keggorg.columns:
                g.annotation['uniprot'] = df_keggorg.loc[g.id, 'UniProt'] if type(df_keggorg.loc[g.id, 'UniProt'])==list else []
                
    logger.info("Added gene annotations retrieved from KEGG.")
                
                
                
            



import sys

import pandas as pnd
import cobra

from .repeating import check_curator
from .repeating import check_rstring_arrow
from .repeating import check_gpr
from .repeating import add_reaction
from .repeating import get_curator_notes


    
def introduce_metabolites(logger, db, model, idcollection_dict, kegg_compound_to_others, outdir, goodbefore, onlycurator):
    goodbefore_reached = False
    
    
    logger.info("Parsing metabolites ('M' sheet)...")

    
    # check duplicated puremids:
    if len(set(db['M']['pure_mid'].to_list())) != len(db['M']): 
        pure_mids = db['M']['pure_mid'].to_list()
        duplicates = list(set([item for item in pure_mids if pure_mids.count(item) > 1]))
        logger.error(f"Sheet 'M' has duplicated metabolites: {duplicates}.")
        return 1
   
        
    # parse M (row by row):    
    db['M'] = db['M'].set_index('pure_mid', drop=True, verify_integrity=True)
    kc_ids_modeled = set()   # account for kc codes modeled
    cnt = 0  # counter for parsed records
    msg = '' # to be cleared
    for iteration, (pure_mid, row) in enumerate(db['M'].iterrows()):
        
        
        # skip empty lines!
        if type(pure_mid) != str: continue
        if pure_mid.strip() == '': continue
        if pure_mid == goodbefore:
            goodbefore_reached = True
            
            
        # manage goodbefore/onlycurator
        if goodbefore != None and goodbefore_reached:
            if onlycurator == None:
                logger.warning(f"Skipping metabolite '{pure_mid}' as requested with --goodbefore[0] '{goodbefore}'.")
                continue
        
        
        # parse and get curators
        response = check_curator(logger, pure_mid, row, db, 'M')
        if type(response) == int: return 1
        else: curators = response
        
        
        # manage goodbefore/onlycurator
        if goodbefore != None and goodbefore_reached:
            if onlycurator != None and onlycurator not in curators:
                curators_string = '; '.join(curators)
                logger.warning(f"Skipping metabolite '{pure_mid}' (curators '{curators_string}') as requested with --goodbefore[0] '{goodbefore}' and --onlycurator '{onlycurator}'.")
                continue
            
        
        # parse formula:
        if pnd.isna(row['formula']):
            logger.error(f"Metabolite '{pure_mid}' has missing formula: '{row['formula']}'.")
            return 1
  
        
        # parse charge: 
        if pnd.isna(row['charge']): 
            logger.error(f"Metabolite '{pure_mid}' has missing charge: '{row['charge']}'.")
            return 1
        
        
        # parse kc:
        if pnd.isna(row['kc']): 
            logger.error(f"Metabolite '{pure_mid}' has missing KEGG annotation (kc): '{row['kc']}'.")
            return 1
        kc_ids = row['kc'].split(';')
        kc_ids = [i.strip() for i in kc_ids]
        for kc_id in kc_ids:
            if kc_id == 'CXXXXX':  # not in KEGG; could be knowledge gap (e.g. methyl group acceptor in R10404)
                with open(f"{outdir}/logs/M.notkegg.txt", 'a') as f:
                    print(f"Metabolite '{pure_mid}' is not in KEGG ('{kc_id}')!", file=f)
                continue  
            #    
            # check if 'kc' codes are real:
            if kc_id not in idcollection_dict['kc']:
                logger.error(f"Metabolite '{pure_mid}' has invalid KEGG annotation (kc): '{kc_id}'.")
                return 1
            #
            # check if 'kc' was already used:
            if kc_id in kc_ids_modeled:
                logger.error(f"KEGG annotation (kc) '{kc_id}' used in metabolite '{pure_mid}' is duplicated.")
                return 1
            if kc_id != 'CXXXXX':
                kc_ids_modeled.add(kc_id)
            
            
        # check the existance of the inchikey
        if pnd.isna(row['inchikey']): 
            logger.error(f"Metabolite '{pure_mid}' has missing inchikey: '{row['inchikey']}'.")
            return 1
        
        # check inchikey format:
        if len(row['inchikey']) != 27 or row['inchikey'][14] != '-' or row['inchikey'][25] != '-':
            logger.error(f"Metabolite '{pure_mid}' has badly formatted inchikey: '{row['inchikey']}'.")
            return 1
        
        
        # check if this 'kc' is already in BiGG (rely on MNX)
        eqbiggids = set()
        for kc_id in kc_ids:
            if kc_id != 'CXXXXX':
                if kc_id in kegg_compound_to_others.keys():
                    for eqbiggid in kegg_compound_to_others[kc_id]['bigg.metabolite']:
                        eqbiggids.add(eqbiggid)
        if pure_mid not in eqbiggids and eqbiggids != set():
            with open(f"{outdir}/logs/M.inbigg.txt", 'a') as f:
                print(f"Metabolites '{'; '.join(kc_ids)}' already in BiGG as {eqbiggids} ({curators} gave '{pure_mid}').", file=f)        
        
        
        # add metabolite to model
        m = cobra.Metabolite(f'{pure_mid}_c')
        model.add_metabolites([m])
        m = model.metabolites.get_by_id(f'{pure_mid}_c')
        m.name = row['name'].strip()
        m.formula = row['formula']
        m.charge = row['charge']
        m.compartment='c'
        
        
        # write curators as annotations
        m.annotation['curator_codes'] = curators
        
        
        # add annotations to model (same order of Memote)
        ankeys = [
            'pubchem.compound', 'kegg.compound', 'seed.compound',
            'inchikey', 'inchi', 'chebi', 'hmdb', 'reactome',
            'metanetx.chemical', 'bigg.metabolite', 'biocyc',
        ]
        # initialize sets:
        for ankey in ankeys:
            if ankey == 'kegg.compound': m.annotation[ankey] = set(kc_ids) - set(['CXXXXX'])
            else: m.annotation[ankey] = set()
        #    
        # populate sets:
        for kc_id in kc_ids:
            if kc_id != 'CXXXXX':
                if kc_id in kegg_compound_to_others.keys():
                    for ankey in ankeys:
                        m.annotation[ankey].update(kegg_compound_to_others[kc_id][ankey])
        #
        # save as list: 
        for ankey in ankeys:
            m.annotation[ankey] = list(m.annotation[ankey])
            
            
        # # force the manual-curated version of the inchikey
        if m.annotation['inchikey'] != [] and m.annotation['inchikey'] != [row['inchikey']]:
            with open(f"{outdir}/logs/M.diffinchi.txt", 'a') as f:
                print(f"Metabolite '{pure_mid}': manual-curated inchikey ({[row['inchikey']]}) is different from the one derived from MNX ({m.annotation['inchikey']}).", file=f)
        m.annotation['inchikey'] = [row['inchikey']]  
        #
        # remove inchikey if unknown:
        if m.annotation['inchikey'] == ['XXXXXXXXXXXXXX-XXXXXXXXXX-X']:
            m.annotation['inchikey'] = []
            
            
        # add SBO annotation
        m.annotation['sbo'] = ['SBO:0000247']  # generic metabolite
        
        
        # add curator_notes
        m.annotation['curator_notes'] = get_curator_notes(logger, row)
        
        
        # communicate progress:
        cnt += 1
        msg = f"Done {cnt}/{len(db['M'])}!"
        print(msg, file=sys.stderr, end='\r')
    print(''.join([' ' for i in range(len(msg))]), file=sys.stderr, end='\r')
   
        
    # check goodbefore reaching:
    if goodbefore != None and goodbefore_reached == False:
        logger.warning(f"Metabolite '{goodbefore}' never reached. Are you sure about your --goodbefore?")
                    
                    
    return model
    
    
    
def introduce_reactions(logger, db, model, idcollection_dict, kegg_reaction_to_others, outdir, goodbefore, onlycurator): 
    goodbefore_reached = False
    
    
    logger.info("Parsing non-transport reactions ('R' sheet)...")
                    
    
    # check duplicated rids:
    if len(set(db['R']['rid'].to_list())) != len(db['R']): 
        rids = db['R']['rid'].to_list()
        duplicates = list(set([item for item in rids if rids.count(item) > 1]))
        logger.error(f"Sheet 'R' has duplicated reactions: {duplicates}.")
        return 1
    
        
    # parse R (row by row):
    db['R'] = db['R'].set_index('rid', drop=True, verify_integrity=True)
    cnt = 0  # counter for parsed records
    msg = '' # to be cleared
    for rid, row in db['R'].iterrows():
        
        
        # skip empty lines!
        if type(rid) != str: continue
        if rid.strip() == '': continue
        if rid == goodbefore:
            goodbefore_reached = True
        
        
        # manage goodbefore/onlycurator
        if goodbefore != None and goodbefore_reached:
            if onlycurator == None:
                logger.warning(f"Skipping reaction '{rid}' as requested with --goodbefore[1] '{goodbefore}'.")
                continue
        
        
        # parse and get curators
        response = check_curator(logger, rid, row, db, 'R')
        if type(response) == int: return 1
        else: curators = response
                    
                    
        # manage goodbefore/onlycurator
        if goodbefore != None and goodbefore_reached:
            if onlycurator != None and onlycurator not in curators:
                curators_string = '; '.join(curators)
                logger.warning(f"Skipping reaction '{rid}' (curators '{curators_string}') as requested with --goodbefore[1] '{goodbefore}' and --onlycurator '{onlycurator}'.")
                continue
        
        
        # parse reaction string
        response = check_rstring_arrow(logger, rid, row, 'R')
        if response == 1: return 1
        

        # parse 'kr':
        if pnd.isna(row['kr']): 
            logger.error(f"Reaction '{rid}' has missing KEGG annotation (kr): '{row['kr']}'.")
            return 1
        kr_ids = row['kr'].split(';')
        kr_ids = [i.strip() for i in kr_ids]
        for kr_id in kr_ids:
            if kr_id == 'RXXXXX':  # not in KEGG; could be knowledge gap 
                with open(f"{outdir}/logs/R.notkegg.txt", 'a') as f:
                    print(f"Reaction '{rid}' is not in KEGG ('{kr_id}')!", file=f)
                continue  
            #
            # check if 'kr' codes are real:
            if kr_id not in idcollection_dict['kr']:
                logger.error(f"Reaction '{rid}' has invalid KEGG annotation (kr): '{kr_id}'.")
                return 1
        
            
        # check GPR:
        response = check_gpr(logger, rid, row, kr_ids, idcollection_dict, 'R', outdir)
        if response == 1: return 1
    
    
        # check if this 'kr' is already in BiGG (rely on MNX)
        eqbiggids = set()
        for kr_id in kr_ids:
            if kr_id != 'RXXXXX':
                if kr_id in kegg_reaction_to_others.keys():
                    for eqbiggid in kegg_reaction_to_others[kr_id]['bigg.reaction']:
                        eqbiggids.add(eqbiggid)
        if rid not in eqbiggids and eqbiggids != set():
            with open(f"{outdir}/logs/R.inbigg.txt", 'a') as f:
                print(f"Reactions '{'; '.join(kr_ids)}' already in BiGG as {eqbiggids} ({curators} gave '{rid}').", file=f) 
        
        
        # add reaction to model
        response = add_reaction(logger, model, rid, curators, row, kr_ids, kegg_reaction_to_others, 'R', outdir)
        if response == 1: return 1
    
    
        # communicate progress:
        cnt += 1
        msg = f"Done {cnt}/{len(db['R'])}!"
        print(msg, file=sys.stderr, end='\r')
    print(''.join([' ' for i in range(len(msg))]), file=sys.stderr, end='\r')
               
    
    # check goodbefore reaching:
    if goodbefore != None and goodbefore_reached == False:
        logger.warning(f"Reaction '{goodbefore}' never reached. Are you sure about your --goodbefore?")
    
    
    return model
      
    
    
def introduce_transporters(logger, db, model, idcollection_dict, kegg_reaction_to_others, outdir, goodbefore, onlycurator): 
    goodbefore_reached = False
    
    
    logger.info("Parsing transport reactions ('T' sheet)...")
    
    
    def clone_to_external(model, mid_c, mid_e):
        # given an existing '_c' M, create its '_e' equivalent
    
        m = cobra.Metabolite(f'{mid_e}')
        model.add_metabolites([m])
        
        m_c = model.metabolites.get_by_id(f'{mid_c}')
        m_e = model.metabolites.get_by_id(f'{mid_e}')
        m_e.compartment='e'
        
        m_e.name = m_c.name
        m_e.formula = m_c.formula
        m_e.charge = m_c.charge
        
        m_e.annotation = m_c.annotation   # transfer all annotations, including SBO!
            
    
    def add_exchange_reaction(model, mid_e):
        # given an existing '_e' M, create the corresponding EX_change reaction
        
        r = cobra.Reaction(f'EX_{mid_e}')
        model.add_reactions([r])
        r = model.reactions.get_by_id(f'EX_{mid_e}')
        
        r.name = f"Exchange for {model.metabolites.get_by_id(mid_e).name}"
        r.build_reaction_from_string(f'{mid_e} --> ')
        r.bounds = (0, 1000)
            
        # add SBO annotation
        r.annotation['sbo'] = ['SBO:0000627']  # exchange reaction    
        
    
    
    # get all already inserted metabolites and reactions
    mids_parsed = [m.id for m in model.metabolites]
    rids_parsed = [r.id for r in model.reactions]
    
    
    # parse T (row by row):
    db['T'] = db['T'].set_index('rid', drop=True, verify_integrity=True)
    cnt = 0  # counter for parsed records
    msg = '' # to be cleared
    for rid, row in db['T'].iterrows():
        
        
        # skip empty lines!
        if type(rid) != str: continue
        if rid.strip() == '': continue
        if rid == goodbefore:
            goodbefore_reached = True
            
            
        # avoid duplicates!
        if rid in rids_parsed:
            logger.error(f"Tranport '{rid}' has ID identical to previously added reaction!")
            return 1
            
            
        # manage goodbefore/onlycurator
        if goodbefore != None and goodbefore_reached:
            if onlycurator == None:
                logger.warning(f"Skipping transport '{rid}' as requested with --goodbefore[2] '{goodbefore}'.")
                continue
        
        
        # parse curator
        response = check_curator(logger, rid, row, db, 'T')
        if type(response) == int: return 1
        else: curators = response
                    
                    
        # manage goodbefore/onlycurator
        if goodbefore != None and goodbefore_reached:
            if onlycurator != None and onlycurator not in curators:
                curators_string = '; '.join(curators)
                logger.warning(f"Skipping transport '{rid}' (curators '{curators_string}') as requested with --goodbefore[2] '{goodbefore}' and --onlycurator '{onlycurator}'.")
                continue
        
        
        # parse reaction string
        response = check_rstring_arrow(logger, rid, row, 'T')
        if response == 1: return 1
    
    
        # parse 'kr':
        if pnd.isna(row['kr']): 
            logger.error(f"Reaction '{rid}' has missing KEGG annotation (kr): '{row['kr']}'.")
            return 1
        if row['kr'] != '-':
            kr_ids = row['kr'].split(';')
            kr_ids = [i.strip() for i in kr_ids]
            for kr_id in kr_ids:
                #
                # check if 'kr' codes are real
                if kr_id not in idcollection_dict['kr']:
                    logger.error(f"Reaction '{rid}' has invalid KEGG annotation (kr): '{kr_id}'.")
                    return 1
        else: 
            # no 'kr' for the majority of transport reactions!
            kr_ids = []

            
        # check GPR:
        response = check_gpr(logger, rid, row, kr_ids, idcollection_dict, 'T', outdir)
        if response == 1: return 1
        
        
        # iterate the involved metabolites
        involved_mids = row['rstring'].split(' ')  # dirty (arrows, coefficints are included)
        for mid in involved_mids: 
            if mid.endswith('_e'):
                mid_e = mid
                mid_c = mid.rsplit('_', 1)[0] + '_c'
                #
                # the cytosolic counterpart must be already modeled:
                if mid_c not in mids_parsed:
                    logger.error(f"{rid}: the metabolite '{mid_c}', counterpart of '{mid_e}', was not previously modeled.")
                    return 1
                #
                # clone to add external metabolite to model
                if mid_e not in mids_parsed:
                    clone_to_external(model, mid_c, mid_e)
                    mids_parsed.append(mid_e)
                #    
                # add corresponding exchange reaction to model
                if f'EX_{mid_e}' not in rids_parsed:
                    add_exchange_reaction(model, mid_e)
                    rids_parsed.append(f'EX_{mid_e}')
                    
                    
        # check if this 'kr' is already in BiGG (rely on MNX)
        eqbiggids = set()
        for kr_id in kr_ids:
            if kr_id != '-':  # (was 'RXXXXX' for metabolic reactions)
                if kr_id in kegg_reaction_to_others.keys():
                    for eqbiggid in kegg_reaction_to_others[kr_id]['bigg.reaction']:
                        eqbiggids.add(eqbiggid)
        if rid not in eqbiggids and eqbiggids != set():
            with open(f"{outdir}/logs/T.inbigg.txt", 'a') as f:
                print(f"Reactions '{'; '.join(kr_ids)}' already in BiGG as {eqbiggids} ({curators} gave '{rid}').", file=f) 
                    
                    
        # add reaction to model
        response = add_reaction(logger, model, rid, curators, row, kr_ids, kegg_reaction_to_others, 'T', outdir)
        if response == 1: return 1
        rids_parsed.append(rid)   # update list of rids in model
        
        
        # communicate progress:
        cnt += 1
        msg = f"Done {cnt}/{len(db['T'])}!"
        print(msg, file=sys.stderr, end='\r')
    print(''.join([' ' for i in range(len(msg))]), file=sys.stderr, end='\r')
        
        
    # check goodbefore reaching:
    if goodbefore != None and goodbefore_reached == False:
        logger.warning(f"Transport '{goodbefore}' never reached. Are you sure about your --goodbefore?")
        
    
    return model



def introduce_sinks_demands(logger, model, dbuni): 
    
    
    logger.debug("Introducing sinks and demands...")
    
    
    # iterate table rows:
    modeled_mids = [m.id for m in model.metabolites]
    A = dbuni['A'].set_index('pure_mid', drop=True, verify_integrity=True)
    sinks = []
    demands = []
    for puremid, row in A.iterrows():
        typee = row['type']  # type is reserved 
        
        
        # check type
        if typee not in ['sink', 'demand']:
            logger.error(f"Unrecognized type '{typee}' for '{puremid}'.")
            return 1
        
        
        # check if precursors are modeled in the universe:
        if f"{puremid}_c" not in modeled_mids:
            logger.error(f"Metabolite '{puremid}', indicated as '{typee}', is not modeled.")
            return 1
        
        
        # populate lists
        if typee == 'sink': sinks.append(puremid)
        else: demands.append(puremid)
    
    
    # introduce sinks 
    for puremid in sinks: 
        r = cobra.Reaction(f'sn_{puremid}_c')
        model.add_reactions([r])
        r = model.reactions.get_by_id(f'sn_{puremid}_c')
        r.name = f"Sink for {model.metabolites.get_by_id(f'{puremid}_c').name}"
        r.build_reaction_from_string(f'{puremid}_c <=> ')
        r.bounds = (-1000, 1000)
        
        
        # add SBO annotation
        r.annotation['sbo'] = ['SBO:0000632']  # sink reaction 
    
    
    # introduce demands
    for puremid in demands: 
        r = cobra.Reaction(f'dm_{puremid}_c')
        model.add_reactions([r])
        r = model.reactions.get_by_id(f'dm_{puremid}_c')
        r.name = f"Demand for {model.metabolites.get_by_id(f'{puremid}_c').name}"
        r.build_reaction_from_string(f'{puremid}_c --> ')
        r.bounds = (0, 1000)
        
        
        # add SBO annotation
        r.annotation['sbo'] = ['SBO:0000628']  # demand reaction 
        
        
    return model
    
    



        

    



import warnings
import logging
import threading
import zipfile
from pathlib import Path
import tempfile
import json
import os
import shutil

import cobra
from escher import Builder

from .downloads import SimpleLoadingWheel



def print_json_tree(data, level=0, max_level=2):
    # explore contents of a json object
    
    if level > max_level:
        return
    indent = '  ' * level
    if isinstance(data, dict):
        for key, value in data.items():
            print(f"{indent}{key}")
            print_tree(value, level + 1, max_level)
    elif isinstance(data, list):
        for i, item in enumerate(data):
            print(f"{indent}[{i}]")
            print_tree(item, level + 1, max_level)
            
            
            
def subset_for_focus(universe, rids_in_group, outdir, focus):
            
    universe_focus = universe.copy()
    to_remove = [r for r in universe_focus.reactions if r.id not in rids_in_group]


    # trick to avoid the WARNING "cobra/core/group.py:147: UserWarning: need to pass in a list" 
    # triggered when trying to remove reactions that are included in groups. 
    with warnings.catch_warnings():  # temporarily suppress warnings for this block
        warnings.simplefilter("ignore")  # ignore all warnings
        cobra_logger = logging.getLogger("cobra.util.solver")
        old_level = cobra_logger.level
        cobra_logger.setLevel(logging.ERROR)   

        universe_focus.remove_reactions(to_remove, remove_orphans=True)

        # restore original behaviour: 
        cobra_logger.setLevel(old_level)  


    # save the subset for drawing in Escher!
    cobra.io.save_json_model(universe_focus, f'{outdir}/focus_{focus}.json')
            
            
        
def count_undrawn_rids_focus(logger, universe, lastmap, focus, outdir):
    
    
    # there could be no tracked folder / no versions for this group
    if lastmap == None:
        return
    
    
    # get modeled reads for this --focus:
    rids_in_group = set()
    try: gr = universe.groups.get_by_id(focus)
    except: 
        logger.warning(f"Group '{focus}' not found!")
        return 
    for r in gr.members:
        rids_in_group.add(r.id)


    # get rids on Escher: 
    drawn_rids = set()
    for key, value in lastmap['json'][1]['reactions'].items():
        drawn_rids.add(value['bigg_id'])

        
    # get remaining rids for this map:
    remainings = rids_in_group - drawn_rids
    remainings_krs = set()
    for rid in remainings: 
        r = universe.reactions.get_by_id(rid)
        if 'kegg.reaction' in r.annotation.keys():
            krs = r.annotation['kegg.reaction']
            for kr in krs: 
                remainings_krs.add(kr)
    
    
    if len(remainings) > 0:
        if focus != 'gr_transport':
            logger.warning(f"Current '{lastmap['filename']}' is {len(remainings)} reactions behind: {' '.join(list(remainings_krs))}.")
        else:
            logger.warning(f"Current '{lastmap['filename']}' is {len(remainings)} reactions behind.")  # usually no kegg codes for tranport reactions
        
        
        # subset the universe to ease the drawing: 
        next_version_filename = f"{focus}-v{(int(lastmap['filename'].rsplit('-v',1)[-1].replace('.json', ''))+1)}.json"
        logger.warning(f"Writing model '{outdir}/focus_{focus}.json' to ease the drawing of '{next_version_filename}'...")
        
        
        t1 = threading.Thread(target = subset_for_focus, args=(
            universe, rids_in_group, outdir, focus))
        t1.start()
        slw = SimpleLoadingWheel(msg="Please wait... ")
        while t1.is_alive():
            slw.proceed()
        slw.clear()
        
        
        logger.warning(f"'{outdir}/focus_{focus}.json' created!")
    else:
        logger.info(f"Current '{lastmap['filename']}' is 0 reactions behind. Thank you ♥")
        
        
        
def parse_zipped_escher(logger, universe, escherzip, outdir): 
    # used to parse the zipped 'escher' folder downloaded from Google Drive.
    # 'escherzip' is the path to zipped folder. 
    
    
    logger.info("Processing collection of hand-drawn Escher maps...")
    
    
    # prepare empty logs folder
    shutil.rmtree(f'{outdir}/escherapps', ignore_errors=True)
    os.makedirs(f'{outdir}/escherapps', exist_ok=True)
    
    
    drawngid_to_maps = dict()
    drawnrid_to_maps = dict()
    drawnpuremid_to_maps = dict()
    with tempfile.TemporaryDirectory() as tmp_dir:
        
        # check path existance: 
        zip_path = Path(escherzip)  # convert to Path for convenience
        if not zip_path.exists():
            logger.error(f"Zipped escher does not exist at provided path '{zip_path}'.")
            return 1
        
        # extract in a auto-deleting temporary directory 
        with zipfile.ZipFile(zip_path, "r") as z:
            z.extractall(tmp_dir)

        # extract the inner child dirs:
        tmp_dir = Path(tmp_dir)
        child_dirs = [p for p in tmp_dir.iterdir() if p.is_dir()]
        escher_dir = child_dirs[0]  # the top-level folder in the ZIP
        inner_child_dirs = [p for p in escher_dir.iterdir() if p.is_dir()]
        
        # iterate the inner child dirs:
        for map_dir in inner_child_dirs:
            map_id = map_dir.name.rsplit('/',1)[-1]
            map_files = [f for f in map_dir.iterdir() if f.is_file()]
            map_files_name = set([f.name.rsplit('/',1)[-1].split('-v',1)[0] for f in map_files])
            
            # check presence of maps:
            if len(map_files_name) == 0:
                # still no maps in this folder
                continue
            
            # check name consistency:
            if len(map_files_name) != 1:
                logger.error(f"Name inconsistency in folder '{map_dir_name}'!")
                return 1
            
            # check consistency with the folder 
            if list(map_files_name)[0] != map_id:
                logger.error(f"Versions in '{map_id}' refer to a different map!")
            
            # get latest version
            map_versions = [int(f.name.rsplit('/',1)[-1].split('-v',1)[-1].replace('.json','')) for f in map_files]
            latest_version = max(map_versions)
            
            
            # read json (last version) to fill the dict
            latest_filepath = str(map_dir / f"{map_id}-v{latest_version}.json")
            with open(latest_filepath, 'r') as file:
                json_data = json.load(file)
        
                # get elements on Escher:
                map_drawngids = set()
                map_drawnrids = set()
                map_drawnpuremids = set()
                for key, value in json_data[1]['reactions'].items():
                    map_drawnrids.add(value['bigg_id'])
                    for i in value['genes']:
                        map_drawngids.add(i['bigg_id'])
                    #for i in value['metabolites']:
                    #    puremeid = i['bigg_id'].rsplit('_',1)[0]
                    #    map_drawnpuremids.add(puremeid)
                for key, value in json_data[1]['nodes'].items():
                    if value['node_type'] == 'metabolite':
                        puremeid = value['bigg_id'].rsplit('_',1)[0]
                        map_drawnpuremids.add(puremeid)
                    
                # populat dicts
                for drawngid in map_drawngids:
                    if drawngid not in drawngid_to_maps.keys():
                        drawngid_to_maps[drawngid] = set()
                    drawngid_to_maps[drawngid].add(map_id)
                for drawnrid in map_drawnrids:
                    if drawnrid not in drawnrid_to_maps.keys():
                        drawnrid_to_maps[drawnrid] = set()
                    drawnrid_to_maps[drawnrid].add(map_id)
                for drawnpuremid in map_drawnpuremids:
                    if drawnpuremid not in drawnpuremid_to_maps.keys():
                        drawnpuremid_to_maps[drawnpuremid] = set()
                    drawnpuremid_to_maps[drawnpuremid].add(map_id)
                    
                    
            # read json (last version) to create the escher app
            builder = Builder(
                map_json = latest_filepath,   
                model_json = None, 
            )
            builder.never_ask_before_quit = True
            builder.scroll_behavior = 'zoom'

            builder.menu = 'all'
            builder.enable_editing = False

            builder.enable_keys = False  # switch not working
            builder.enable_search = True  # switch not working
            builder.full_screen_button = True  # switch not working

            builder.highlight_missing = True 
            builder.show_gene_reaction_rules = True
            builder.enable_tooltips = 'label'
    
            builder.save_html(f"{outdir}/escherapps/{map_id}.html")
            
        
    # apply annotation to universe
    for g in universe.genes: 
        if g.id in drawngid_to_maps.keys():
            g.annotation['drawn_in_maps'] = list(drawngid_to_maps[g.id])
        else:
            g.annotation['drawn_in_maps'] = []
    for r in universe.reactions: 
        if r.id in drawnrid_to_maps.keys():
            r.annotation['drawn_in_maps'] = list(drawnrid_to_maps[r.id])
        else:
            r.annotation['drawn_in_maps'] = []
    for m in universe.metabolites: 
        puremid = m.id.rsplit('_',1)[0]
        if puremid in drawnpuremid_to_maps.keys():
            m.annotation['drawn_in_maps'] = list(drawnpuremid_to_maps[puremid])
        else:
            m.annotation['drawn_in_maps'] = []
            
    
    return universe


import warnings
from decimal import Decimal

import numpy as np
import pandas as pnd
import cobra

from .coeffs import get_uni_biom_coeff
from .coeffs import check_exp_biomass_data
from .coeffs import compute_exp_DNA_coeffs
from .coeffs import compute_exp_RNA_coeffs
from .coeffs import compute_exp_PROTS_coeffs
from .coeffs import compute_exp_LIPIDS_coeffs
from .coeffs import compute_exp_WTA_coeffs
from .coeffs import compute_exp_LTA_coeffs
from .coeffs import compute_exp_PEPTIDO_coeffs
from .coeffs import compute_exp_EPS_coeffs



def get_biomass_dict(logger, universe, dbuni):
    
    
    # iterate table rows:
    universal_mids = [m.id for m in universe.metabolites]
    B = dbuni['B'].set_index('pure_mid', drop=True, verify_integrity=True)
    class_to_precursors = dict()
    for puremid, row in B.iterrows():
        classs = row['class']  # class is reserved 
        
        
        # check if precursors are modeled in the universe:
        if f"{puremid}_c" not in universal_mids:
            logger.error(f"Metabolite '{puremid}', indicated in '{classs}', is not present in universe.")
            return 1
        
        
        # populate dic
        if classs not in class_to_precursors.keys():
            class_to_precursors[classs] = set()
        class_to_precursors[classs].add(f"{puremid}_c")
    
                                        
    # Cofactors:
    # Note: universal ('cofs_uni_Xavier2017') and conditional ('cofs_cond_Xavier2017') cofactors have been defined in Xavier2017 .
    # Here presented in the same order of Xavier2017.
    # Cofactors not mentioned in Xavier2017 ('cofactors_add') are reported below.
    # Xavier2017: 10.1016/j.ymben.2016.12.002.
    
    # 'f4200_c' is coenzyme f420, electron transfer in methanogens, actinobacteria , and others-
    # 'f4200_c', # B2: included by Xavier2017 in 'universal' but the description seems conditional.
    # 'ACP_c', acyl-carrier protein (removed as it contains the 'X' atom, causing error in Memote)                                    
                                        
    return class_to_precursors



def get_precursors_to_pathway(logger, dbuni):
    
    
    B = dbuni['B'].set_index('pure_mid', drop=True, verify_integrity=True)
    precursors_to_pathway = dict()
    for puremid, row in B.iterrows():
        
        
        # check evaluation aprameter
        evaluation = row['evaluation']  # class is reserved 
        allowded = ['universal', 'experimental', 'completeness']
        if evaluation not in allowded:
            logger.error(f"Unrecognized evaluation field for '{puremid}'. Alloweded values are {allowded}.")
            return 1
        
        
        # parse only if 'completeness'
        if evaluation == 'universal': continue
        if evaluation == 'experimental': continue
        
        
        # populate dict
        groups = [i.strip() for i in row['bs_groups'].strip().split(';')]
        precursors_to_pathway[f'{puremid}_c'] = groups
    
    
    return precursors_to_pathway



def introduce_universal_biomass(logger, universe, dbuni): 
    
    
    logger.debug("Introducing universal biomass reaction...")
    
    
    # biold universal biomass reaction string with default coefficients:
    biomass_dict = get_biomass_dict(logger, universe, dbuni)
    rstring = ''
    for classs, precursors in biomass_dict.items():
        rstring = rstring + f'{get_uni_biom_coeff()} {(" + " + str(get_uni_biom_coeff()) + " ").join(precursors)} + '
    rstring = rstring + f'{get_uni_biom_coeff()} atp_c + {get_uni_biom_coeff()} h2o_c --> {get_uni_biom_coeff()} adp_c + {get_uni_biom_coeff()} h_c + {get_uni_biom_coeff()} pi_c'

    
    # convert to a proper cobra.Reaction object
    r = cobra.Reaction('Biomass')
    universe.add_reactions([r])
    r = universe.reactions.get_by_id('Biomass')
    r.name = 'Biomass assembly reaction'
    r.build_reaction_from_string(rstring)
    
    
    # add SBO annotation
    r.annotation['sbo'] = ['SBO:0000629']  # biomass reaction 
    
    
    # set as objective:
    universe.objective = 'Biomass'
    
    
    return universe



def adjust_biomass_precursors(logger, model, universe, conditional_threshold, dbuni):
    
    
    # 'cond_col_dict' is a dictionary of variable biomass precursors
    cond_col_dict = {}
    modeled_rids = [r.id for r in model.reactions]
    cnt_removed = 0

    
    # get and iterate 'precursors_to_pathway'
    precursor_to_pathway = get_precursors_to_pathway(logger, dbuni)
    if type(precursor_to_pathway) == int: return 1
    for precursor, pathways in precursor_to_pathway.items(): 
        
        
        # iterate alternative pathways for this biomass precursor:
        pathway_to_absence = {}
        pathway_to_compstring = {}   # completeness string
        for pathway in pathways:   
            cnt_members_tot = 0
            cnt_members_present = 0

            
            # continue if the pathway is still lmissing form the universe: 
            if pathway not in [gr.id for gr in universe.groups]:
                continue   
                
                
            # count group members in this strain
            for member in universe.groups.get_by_id(pathway).members:
                cnt_members_tot += 1    
                if member.id in modeled_rids:
                    cnt_members_present += 1
                    
                    
            # populate dicts:
            pathway_to_absence[pathway] = (cnt_members_present / cnt_members_tot) < conditional_threshold
            pathway_to_compstring[pathway] = f'{pathway}: {cnt_members_present}/{cnt_members_tot}'
            

        cond_col_dict[precursor] = '; '.join(list(pathway_to_compstring.values()))
        if all(list(pathway_to_absence.values())):
            cnt_removed += 1
            # add metabolites to the right side (they will disappear if the balance is 0)
            model.reactions.Biomass.add_metabolites({precursor: get_uni_biom_coeff()})
            logger.warning(f"Removed biomass precursor '{precursor}' as it seemed not required ({cond_col_dict[precursor]}).")

       
    if model.id != universe.id:   # avoid logging when testing blocked biomass precusors of the universe.
        logger.info(f"Removed {cnt_removed} conditional biomass precursors.")
        
        
    return cond_col_dict



def compute_biomass_norm(logger, model, biomass): 
    
    
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        # avoids warnings like "UserWarning: 9.823 is not an integer (in formula C9.823H5.5789999999999935N3.6769999999999996O3.628P1.0)"
    
    
        norm = Decimal(0.0)
        for m, coeff in model.reactions.get_by_id('Biomass').metabolites.items():
            if coeff < 0.0:   # reactant
                if m.id not in ["atp_c", "h2o_c"]:
                    # '/1000' as they are milli-model coefficients:
                    norm = norm + Decimal(abs(coeff)) / 1000 * Decimal(m.formula_weight) 
        logger.info(f"Biomass adjusted using '{biomass}' data: {str(round(norm, 16))} gDW.")



def adjust_biomass_coefficients(logger, model, universe, dbuni, dbexp, biomass):
    # Strategy: the default coefficients are equal to 'get_uni_biom_coeff()', which is  a really samm number, negligible.
    # Coefficients are adjusted only for precursors comprised by MWF, all the others are left with default coefficients.
    # Therefore, if the final biomass is not sufficiently close to 1.0, just lower the 'get_uni_biom_coeff()', 
    # or lower the number of precursors left with default coefficient. 
    
    
    if biomass == '-':
        logger.info(f"No biomass data provided: using minimal coefficients ({get_uni_biom_coeff()}).")
        return 0
    
    
    # check biomass data
    response = check_exp_biomass_data(logger, dbexp, biomass)
    if type(response)==int: return 1
    else: MWF, DNA, RNA, PROTS, LIPIDS_PL, LIPIDS_FA = response
    
    
    # compute coefficient for 'Biomass' reaction:
    biomass_dict = get_biomass_dict(logger, universe, dbuni)
    reacs_prods_coeffs = dict()
    
    
    # the reaction 'MEANFA' will replace 'sn_meanfa_c' to produce 'meanfa_c'.
    model.remove_reactions([model.reactions.get_by_id("sn_meanfa_c")])
    logger.debug(f"Removed 'sn_meanfa_c': will be replaced by 'MEANFA'.")
    
    
    # DNA
    response = compute_exp_DNA_coeffs(logger, model, MWF, DNA)
    if type(response)==int: return 1
    else: reacs_prods_coeffs['DNA'] = response
    # RNA
    response = compute_exp_RNA_coeffs(logger, model, MWF, RNA)
    if type(response)==int: return 1
    else: reacs_prods_coeffs['RNA'] = response
    # PROTS
    response = compute_exp_PROTS_coeffs(logger, model, MWF, PROTS)
    if type(response)==int: return 1
    else: reacs_prods_coeffs['PROTS'] = response
    # LIPIDS
    response = compute_exp_LIPIDS_coeffs(logger, model, MWF, LIPIDS_PL, LIPIDS_FA)
    if type(response)==int: return 1
    else: reacs_prods_coeffs['LIPIDS'] = response 
    # WTA
    response = compute_exp_WTA_coeffs(logger, model, biomass_dict, MWF, dbexp)    
    if type(response)==int: return 1
    else: reacs_prods_coeffs['WTA'] = response
    # LTA
    response = compute_exp_LTA_coeffs(logger, model, biomass_dict, MWF, dbexp)
    if type(response)==int: return 1
    else: reacs_prods_coeffs['LTA'] = response
    # PEPTIDO
    response = compute_exp_PEPTIDO_coeffs(logger, model, biomass_dict, MWF, dbexp)
    if type(response)==int: return 1
    else: reacs_prods_coeffs['PEPTIDO'] = response
    # EPS
    response = compute_exp_EPS_coeffs(logger, model, biomass_dict, MWF, dbexp)
    if type(response)==int: return 1
    else: reacs_prods_coeffs['EPS'] = response  
    
    
    # merge dicts of reactants and products: 
    reactant_to_coeffs = dict()
    product_to_coeffs = dict()
    for i in reacs_prods_coeffs.values():
        reacs, prods = i[0], i[1]
        for key, coeff in reacs.items():
            if key not in reactant_to_coeffs.keys():
                reactant_to_coeffs[key] = Decimal('0.0')
            reactant_to_coeffs[key] = reactant_to_coeffs[key] + coeff
        for key, coeff in prods.items():
            if key in ['DNA_c', 'RNA_c', 'PROTS_c', 'LIPIDS_c']:
                continue
            if key not in product_to_coeffs.keys():
                product_to_coeffs[key] = Decimal('0.0')
            product_to_coeffs[key] = product_to_coeffs[key] + coeff
    
    
    # update biomass coefficients
    for mid in [m.id for m in model.reactions.get_by_id('Biomass').reactants]:
        if mid in reactant_to_coeffs.keys(): 
            model.reactions.Biomass.add_metabolites({mid: get_uni_biom_coeff()})
            model.reactions.Biomass.add_metabolites({mid: -float(reactant_to_coeffs[mid])})
    for mid in [m.id for m in model.reactions.get_by_id('Biomass').products]:
        if mid in product_to_coeffs.keys(): 
            model.reactions.Biomass.add_metabolites({mid: -get_uni_biom_coeff()})
            model.reactions.Biomass.add_metabolites({mid: float(product_to_coeffs[mid])})
            
            
    # remove unsed biomass precursors:
    biomass_dict = get_biomass_dict(logger, universe, dbuni)
    # remove all fatty acids:
    for mid in biomass_dict["LIPIDS_FA"]:
        model.reactions.Biomass.add_metabolites({mid: get_uni_biom_coeff()})
    # remove unused phospholipids:
    for mid in biomass_dict["LIPIDS_PL"]:
        if mid not in reacs_prods_coeffs['LIPIDS'][0].keys():
            model.reactions.Biomass.add_metabolites({mid: get_uni_biom_coeff()})
            
            
    # check normalization to 1.0 gDW
    compute_biomass_norm(logger, model, biomass)
            
    
    return 0



def get_biomass_df_structure(logger, df, model, universe, dbuni, cond_col_dict):
    
    
    # get strain-specific biomass precursors / coeffs to gap-fill: 
    biomass_mids_to_coeff = dict()
    for m, coeff in model.reactions.Biomass.metabolites.items():
        if coeff < 0:   # reactants
            biomass_mids_to_coeff[m.id] = abs(coeff)
            
            
    # get all universal biomass precursors: 
    biomass_dict = get_biomass_dict(logger, universe, dbuni)
    biomass_mids = []
    for key, values in biomass_dict.items():
        for i in values:
            biomass_mids.append(i)
            
    
    # preassign class to each precursor  (besically invert 'biomass_dict')
    precursor_to_class = dict()
    for classs, mids in biomass_dict.items():  # class is protected keyword
        for mid in mids: 
            precursor_to_class[mid] = classs
    precursor_to_class['atp_c'] = '/'
    
    
    # define universal biomass precursors (atp in 1st position)
    universal_precs = [m.id for m in universe.reactions.Biomass.reactants if m.id not in ['atp_c', 'h2o_c']]
    universal_precs = ['atp_c'] + universal_precs
    
    
    # prepare sheet for excel output
    df['name'] = None
    df['class'] = None
    df['cond'] = None
    df['coeff'] = None
    for mid in universal_precs:
        
        
        # skip if this strain does not produce particular PL/FA 
        if mid in biomass_mids and not mid in biomass_mids_to_coeff.keys():
            if mid in biomass_dict['LIPIDS_PL'] or mid in biomass_dict['LIPIDS_FA']:
                continue    
        
        
        # set base info 
        df.loc[mid, 'name'] = universe.metabolites.get_by_id(mid).name
        df.loc[mid, 'class'] = precursor_to_class[mid]
        
        
        # set 'cond' column:
        if mid in cond_col_dict.keys():
            df.loc[mid, 'cond'] = f'{cond_col_dict[mid]}'
        else:
            df.loc[mid, 'cond'] = '/'
            
            
        # set 'coeff' column:
        if mid not in [m.id for m in model.reactions.Biomass.reactants]:  
                df.loc[mid, 'coeff'] = 'removed'
        else:
            df.loc[mid, 'coeff'] = biomass_mids_to_coeff[mid]
            
            
    return df





import warnings
import logging



def remove_forced(logger, model, universe, force_removal):
    
    
    if force_removal != '-':
        removed_rids = []
        forced_rids = force_removal.split(',')
        
        modeled_rids = [r.id for r in model.reactions]
        universal_rids = [r.id for r in universe.reactions]
        
        for rid in forced_rids: 
            
            if rid not in universal_rids:
                logger.info(f"Ignoring reaction ID '{rid}' since it's not included in the universe.")
                continue
            
            if rid not in modeled_rids:
                logger.debug(f"Requested reaction ID '{rid}' was already excluded from the model.")
            else:
                removed_rids.append(rid)

        # remove collected reactions:
        with warnings.catch_warnings():
            # avoid warnings like "python3.9/site-packages/cobra/core/group.py:147: UserWarning: need to pass in a list"
            model.remove_reactions(removed_rids) # this works also with rids instaead of Rs

        logger.info(f"Reactions forcibly removed: {removed_rids}.")
        


def remove_disconnected(logger, model):
    
    
    # remove metabolites involved in 0 reactions:
    to_remove = []
    for m in model.metabolites:
        if len(m.reactions) == 0:
            to_remove.append(m)
    model.remove_metabolites(to_remove)
    
    
    # return number of removed metabolites
    return len(to_remove)

    
    
def remove_orphans(logger, model, keep_transport):
    
    
    # collect universal orphans
    to_remove = []
    for r in model.reactions:       
        if r.id == 'Biomass': 
            continue   # exclude biomass
        if len(r.metabolites) == 1: 
            continue   # exclude exr, sinks, demands 
        if keep_transport and len(set([m.id.rsplit('_',1)[-1] for m in r.metabolites])) > 1: 
            continue  # exclude transport reactions if requested
        if r.gene_reaction_rule == '': 
            logger.debug(f"Removing orphan '{r.id}'...")
            to_remove.append(r.id)
            
    
    # trick to avoid the WARNING "cobra/core/group.py:147: UserWarning: need to pass in a list" 
    # triggered when trying to remove reactions that are included in groups. 
    with warnings.catch_warnings():  # temporarily suppress warnings for this block
        warnings.simplefilter("ignore")  # ignore all warnings
        cobra_logger = logging.getLogger("cobra.util.solver")
        old_level = cobra_logger.level
        cobra_logger.setLevel(logging.ERROR)   

        model.remove_reactions(to_remove) 

        # restore original behaviour: 
        cobra_logger.setLevel(old_level)   
            

    # return number of removed reactions
    return len(to_remove)



def remove_sinks_demands(logger, model, keep_meanfa=False):
    
    
    # collect universal orphans
    to_remove = []
    for r in model.reactions:
        if r.id == 'Biomass': continue   # exclude biomass
        if len(r.metabolites) != 1: continue  # exclude transport and non-transport reactions
        if len(r.metabolites) == 1 and list(r.metabolites)[0].id.endswith('_e'): continue   # exclude EXR
        if r.id == 'sn_meanfa_c' and keep_meanfa: continue
        if r.gene_reaction_rule == '': 
            # should remain only sinks/demands
            logger.debug(f"Removing sink/demand '{r.id}'...")
            to_remove.append(r.id)
            
    
    # trick to avoid the WARNING "cobra/core/group.py:147: UserWarning: need to pass in a list" 
    # triggered when trying to remove reactions that are included in groups. 
    with warnings.catch_warnings():  # temporarily suppress warnings for this block
        warnings.simplefilter("ignore")  # ignore all warnings
        cobra_logger = logging.getLogger("cobra.util.solver")
        old_level = cobra_logger.level
        cobra_logger.setLevel(logging.ERROR)   

        model.remove_reactions(to_remove) 

        # restore original behaviour: 
        cobra_logger.setLevel(old_level)   
            

    # return number of removed reactions
    return len(to_remove)
    
    
    


import os
from pathlib import Path
import logging
import tempfile
import concurrent.futures as confu 
import sys

import cobra

from .pruner import load_input_universe
from .pruner import prune_universe
from .pruner import load_input_eggnog
from .pruner import load_keggorg_like_eggnog
from .pruner import parse_eggnog
from .pruner import subtract_kos
from .pruner import translate_remaining_kos
from .pruner import restore_gene_annotations
from .pruner import append_keggorg_gene_annots

from .gapfillutils import include_forced

from .gapfill import gapfill_on_media

from .polishing import remove_orphans
from .polishing import remove_sinks_demands
from .polishing import remove_forced
from .polishing import remove_disconnected

from .biologcuration import biolog_on_media

from ..runsims.biosynth import biosynthesis_on_media

from ..runsims.simplegrowth import grow_on_media

from ..parsedb.cycles import verify_egc_all

from ..commons.logutils import get_logger

from ..commons.downloads import get_databases
from ..commons.downloads import eggnog_format_expansion

from ..commons.biomass import adjust_biomass_precursors
from ..commons.biomass import adjust_biomass_coefficients

from ..commons.sbmlutils import force_id_on_sbml

from ..commons.excelhub import write_excel_model
from ..commons.excelhub import comparative_table

from ..commons.metrics import log_metrics
from ..commons.metrics import log_unbalances

from ..commons.keggutils import download_keggorg

from ..commons.medium import initialize_model
from ..commons.medium import reset_growth_env 

from ..commons.memoteutils import get_memote_results_dict



def create_model_incore(params):
    
    
    # load params
    annotation_source, universe, pruned_universe, eggpath, dbuni, dbexp, args, multistrain = params
    
    
    # get the logger:
    logger = get_logger('gsrap_queued', args.verbose)  # loggers can't be pickled!
    
    
    # disable logging if strains are more than 1:
    # only errors will be recorded if multistrain mode
    if multistrain:
        logger.setLevel(logging.ERROR)  



    ###### STRAIN-SPECIFIC PRUNING
    
    # create a copy of the universe and define the model ID
    logger.info("Starting reconstruction from a copy of whole universe...")
    model = universe.copy()  # 'universe' and not 'pruned_universe' by design
    
    
    # assign model ID
    if annotation_source == 'keggorg':
        model.id = args.keggorg
    elif annotation_source == 'eggnog':
        model.id = Path(eggpath).stem 

        
    # remove all sinks and demands, excluding sn_meanfa (removed later).
    n = remove_sinks_demands(logger, model, keep_meanfa=True)  
    logger.info(f"Removed {n} universal sinks or demand reactions.")
    

    # remove all orphans 
    n = remove_orphans(logger, model, keep_transport=False) 
    logger.info(f"Removed {n} universal orphan reactions.")

    
    # get the annotation dictionaries
    if annotation_source == 'keggorg':
        eggnog_style_table = load_keggorg_like_eggnog(logger, args.keggorg, args.outdir)
        logger.info(f"Reading annotation for organism code '{args.keggorg}'...")
    elif annotation_source == 'eggnog':
        eggnog_style_table = load_input_eggnog(logger, eggpath)
        logger.info("Reading provided eggnog-mapper annotation...")
    # get important dictionaries: 'eggnog_ko_to_gids' and 'eggonog_gid_to_kos'
    eggnog_ko_to_gids, eggonog_gid_to_kos = parse_eggnog(eggnog_style_table)  
      

    # prune reactions
    subtract_kos(logger, model, eggnog_ko_to_gids)

    
    # translate KOs to the actual genes
    translate_remaining_kos(logger, model, eggnog_ko_to_gids)
    restore_gene_annotations(logger, model, universe, eggonog_gid_to_kos)
    # insert gene annotation if starting from kegg organisms:
    if annotation_source == 'keggorg': 
        append_keggorg_gene_annots(logger, model, args.keggorg, args.outdir)

    

    ###### GAPFILLING
    
    # force inclusion of reactions prior to gap-filling:  
    include_forced(logger, model, universe, args.include)

    
    # remove missing conditional precursors + get the 'cond_col_dict' dict.
    # 'cond_col_dict' is str-to-str: {'pheme_c': 'M00868: 1/8; M00121: 2/12;', 'hemeO_c': 'gr_HemeO: 0/1'}
    cond_col_dict = adjust_biomass_precursors(logger, model, universe, args.conditional, dbuni) 

    
    # adjust biomass coefficients
    response = adjust_biomass_coefficients(logger, model, universe, dbuni, dbexp, args.biomass)
    if response == 1: return 1


    # gap-fill based on media:
    df_B = gapfill_on_media(logger, model, pruned_universe, universe, dbuni, dbexp, args.gapfill, cond_col_dict)
    if type(df_B)==int: return 1


    # force removal of reactions
    # experimental feature, not public: it's main purpose was to test gap-filling in biolog_on_media().
    setattr(args, 'remove', '-')  
    remove_forced(logger, model, universe, args.remove)

    
    # perform Biolog(R) curation based on media
    df_P = biolog_on_media(logger, model, pruned_universe, universe, dbexp, args.gapfill, args.biolog, args.cnps)
    if type(df_P)==int: return 1

    
    # remove disconnected metabolites
    n = remove_disconnected(logger, model)
    logger.info(f"Removed {n} disconnected metabolites.")

    
    # log metrics
    log_metrics(logger, model)
    log_unbalances(logger, model)


    
    ###### POST-RECONSTCTION
    
    # check erroneous EGCs
    verify_egc_all(logger, model, args.outdir)
    
    
    # check blocked metabolites / dead-ends
    df_S = biosynthesis_on_media(logger, model, dbexp, args.gapfill, args.biosynth)
    if type(df_S)==int: return 1


    # reset growth environment befor saving the model (changed during growth sims)
    reset_growth_env(model)

    
    # initialize model
    response = initialize_model(logger, model, dbexp, args.initialize, args.gapfill)
    if response==1: return 1


    # compute Memote metrics
    memote_results_dict = get_memote_results_dict(logger, model)


    # write the model on disk:
    logger.info("Writing strain-specific model...")
    cobra.io.save_json_model(model, f'{args.outdir}/{model.id}.json')        # JSON
    logger.info(f"'{args.outdir}/{model.id}.json' created!")
    cobra.io.write_sbml_model(model, f'{args.outdir}/{model.id}.xml')        # SBML   # groups are saved only to SBML
    logger.info(f"'{args.outdir}/{model.id}.xml' created!")
    force_id_on_sbml(f'{args.outdir}/{model.id}.xml', model.id)   # force introduction of the 'id=""' field
    sheets_dict = write_excel_model(model, f'{args.outdir}/{model.id}.mkmodel.xlsx', args.nofigs, memote_results_dict, None, df_B, df_P, df_S)  
    logger.info(f"'{args.outdir}/{model.id}.mkmodel.xlsx' created!")
        
        
    return sheets_dict    



def multicore_generation_handling(universe, pruned_universe, args, annotation_source, dbuni, dbexp, multistrain):
    
    
    # base objects
    error_raised = False
    sheets_dicts = []
    executor =  confu.ProcessPoolExecutor(max_workers=args.cores)

    
    # run executor according to the annotation source
    if annotation_source == 'keggorg':
        futures = [executor.submit(create_model_incore, (
            annotation_source, universe, pruned_universe, None, dbuni, dbexp, args, multistrain
        ))]
    elif annotation_source == 'eggnog':
        futures = [executor.submit(create_model_incore, (
            annotation_source, universe, pruned_universe, eggpath, dbuni, dbexp, args, multistrain
        )) for eggpath in args.eggnog]
    
    
    # collect results as the runs complete
    for f in confu.as_completed(futures):
        sheets_dict = f.result() 
        
        
        # error raised in one of the cores:
        if type(sheets_dict) == int: 
            error_raised = True
            break   # a single error from a child process is sufficient to shutdown all the executor
        
        
        # this strain reached the end:
        else:
            if multistrain:
                sheets_dicts.append(sheets_dict)
                print(f"{len(sheets_dicts)}/{len(args.eggnog)} ({int(len(sheets_dicts)/len(args.eggnog)*100)}%) completed!", end='\r', file=sys.stderr)
    
    
    # hide last progress trace ('sheets_dicts' unused if not in multi-strain mode):
    if multistrain and sheets_dicts != []:
        last_trace = f"{len(sheets_dicts)}/{len(args.eggnog)} ({int(len(sheets_dicts)/len(args.eggnog)*100)}%) completed!"
        whitewash = ''.join([' ' for i in range(len(last_trace))])
        print(whitewash, end='\r', file=sys.stderr)   
    
    
    # multiproces part terminated: safely shut down the executor
    executor.shutdown(wait=True)
    
    
    # propagate errors if any:
    if error_raised:
        return 1
    
    
    return sheets_dicts



def main(args, logger):
    
    
    # adjust out folder path                      
    while args.outdir.endswith('/'):              
        args.outdir = args.outdir[:-1]
    os.makedirs(f'{args.outdir}/', exist_ok=True)
    
    
    # check compatibility of input parameters:
    if args.cnps == '-' and args.biolog != '-':
        logger.error("Missing starting C/N/P/S sources: --biolog must be used in conjunction with --cnps.")
        return 1
    if args.transrule not in ['rescue', 'ignore']:
        logger.error("Unrecognized value for --transrule: it must be either 'rescue' or 'ignore'.")
        return 1
    
    
    # determine available cores
    cpu_count = os.cpu_count()
    if cpu_count == None: 
        logger.debug("Failed to count cpus. Reset to 1.")
        cpu_count = 1  
    if args.cores == 0:
        args.cores = cpu_count
    elif args.cores > cpu_count:
        logger.debug(f"Parameter --cores {args.cores} is greater than the maximum available ({cpu_count}): reset to {cpu_count}.")
        args.cores = cpu_count
    
    
    # format the --eggnog param
    args.eggnog = eggnog_format_expansion(logger, args.eggnog)  # now 'args.eggnog' could still be '-'
    
    
    # get the kegg organism if requested
    if args.keggorg != '-':
        response = download_keggorg(logger, args.keggorg, args.outdir)
        if response == 1: return 1
        
    
    # determine the source of functional annotation:
    annotation_source = None
    if args.keggorg != '-':  # keggorg has precedence
        annotation_source = 'keggorg'
    elif args.eggnog != '-':
        annotation_source = 'eggnog'
        if args.cores > len(args.eggnog):
            logger.debug(f"Parameter --cores {args.cores} is greater than the number of strains ({len(args.eggnog)}): reset to {len(args.eggnog)}.")
            args.cores = len(args.eggnog)
    else:
        logger.error("No valid functional annotations provided: please use '--keggorg' or '--eggnog'.")
        return 1
    
    
    # get dbexp (dbuni not really used):
    logger.info("Downloading updated gsrap assets...")
    response = get_databases(logger)
    if type(response)==int: return 1
    else: dbuni, dbexp = response 
    
    
    # load universe (fresh or built-in):
    universe = load_input_universe(logger, args.universe)
    if type(universe)==int:
        return 1
    
    
    # check and use the --taxon param to prune the universe
    pruned_universe = prune_universe(logger, universe, args.taxon, args.transrule)
    if type(pruned_universe) == int: return 1
    
    
    # determine if mkmodel will be run in multistrain mode:
    if annotation_source == 'keggorg':
        multistrain = False
    elif annotation_source == 'eggnog':
        multistrain = len(args.eggnog) > 1
        if multistrain:
            logger.info(f"Number of provided strains is >1: logging will be disabled.")
            logger.info(f"Performing {len(args.eggnog)} reconstructions relying on {args.cores} cores... ")
            
            
    # create strain-specific GSMMs using multi-core
    sheets_dict = multicore_generation_handling(universe, pruned_universe, args, annotation_source, dbuni, dbexp, multistrain)
    if type(sheets_dict) == int: 
        return 1  # some error was raised 
    
        
    # everything went fine, and there are > 1 strains: time to make the comparison table:
    if multistrain:
        logger.info("Collecting comparative tables...")
        comparative_table(logger, args.outdir, sheets_dicts)
            
        
    return 0


import os
import pickle
from importlib import resources
import shutil

import cobra

from .introduce import introduce_metabolites
from .introduce import introduce_reactions
from .introduce import introduce_transporters
from .introduce import introduce_sinks_demands

from .annotation import translate_annotate_genes
from .annotation import set_up_groups

from .completeness import check_completeness

from .cycles import verify_egc_all

from ..runsims.simplegrowth import grow_on_media

from ..runsims.precursors import precursors_on_media

from ..runsims.biosynth import biosynthesis_on_media

from ..commons.sbmlutils import force_id_on_sbml

from ..commons.metrics import log_metrics
from ..commons.metrics import log_unbalances
from ..commons.metrics import show_contributions

from ..commons.downloads import get_databases
from ..commons.downloads import eggnog_format_expansion
from ..commons.downloads import check_taxon
from ..commons.downloads import get_eschermap

from ..commons.biomass import introduce_universal_biomass
from ..commons.biomass import adjust_biomass_precursors

from ..commons.excelhub import write_excel_model

from ..commons.escherutils import count_undrawn_rids_focus
from ..commons.escherutils import parse_zipped_escher

from ..commons.keggutils import download_keggorg

from ..commons.medium import initialize_model
from ..commons.medium import reset_growth_env

from ..commons.memoteutils import get_memote_results_dict



def main(args, logger):
    

    
    ###### PRE-PARSING
    
    # adjust out folder path                      
    while args.outdir.endswith('/'):              
        args.outdir = args.outdir[:-1]
    os.makedirs(f'{args.outdir}/', exist_ok=True)
    
    
    # prepare empty logs folder
    shutil.rmtree(f'{args.outdir}/logs', ignore_errors=True)
    os.makedirs(f'{args.outdir}/logs', exist_ok=True)
    
    
    # check compatibility of input parameters
    if args.progress==False and args.module==True: 
        logger.error(f"You cannot ask --module without --progress (see --help).")
        return 1
    if args.progress==False and args.focus!='-':
        logger.error(f"You cannot ask --focus without --progress (see --help).")
        return 1
    
    
    # check 'goodbefore' and 'onlycurator' params
    if args.goodbefore == '-' and args.onlycurator != '-':
        logger.error(f"--onlycurator must be used in conjunction with --goodbefore.")
        return 1
    if   args.goodbefore == '-': args.goodbefore = [None, None, None]
    elif len(args.goodbefore.split('-')) != 3: 
        logger.error(f"Invalid syntax detected for --goodbefore.")
        return 1
    else:
        args.goodbefore = args.goodbefore.split('-')
        if args.goodbefore[0] == 'None': args.goodbefore[0] = None
        if args.goodbefore[1] == 'None': args.goodbefore[1] = None
        if args.goodbefore[2] == 'None': args.goodbefore[2] = None
    if args.onlycurator == '-': args.onlycurator = None
    
    
    # check and extract the required 'gsrap.maps' file
    if os.path.exists(f'{args.inmaps}') == False:
        logger.error(f"File 'gsrap.maps' not found at {args.inmaps}.")
        return 1
    try: 
        with open(f'{args.inmaps}', 'rb') as f:
            inmaps = pickle.load(f)  
    except: 
        logger.error(f"Provided file {args.inmaps} has an incorrect format.")
        return 1
    idcollection_dict = inmaps['idcollection_dict']
    summary_dict = inmaps['summary_dict']
    
    
    # load internal resources
    with resources.path("gsrap.assets", f"kegg_compound_to_others.pickle") as asset_path: 
        with open(asset_path, 'rb') as handle:
            kegg_compound_to_others = pickle.load(handle)
    with resources.path("gsrap.assets", f"kegg_reaction_to_others.pickle") as asset_path: 
        with open(asset_path, 'rb') as handle:
            kegg_reaction_to_others = pickle.load(handle) 
            
                        
    # format the --eggnog param
    args.eggnog = eggnog_format_expansion(logger, args.eggnog)  # now 'args.eggnog' could still be '-'
    
    
    # check the --taxon param
    if args.taxon != '-':
        response = check_taxon(logger, args.taxon, idcollection_dict)
        if response == 1: return 1
    
    
    # get the kegg organism if requested
    if args.keggorg != '-':
        response = download_keggorg(logger, args.keggorg, args.outdir)
        if response == 1: return 1
    
    
    # download dbuni, dbexp:
    logger.info("Downloading updated gsrap assets...")
    response = get_databases(logger)
    if type(response)==int: return 1
    else: dbuni, dbexp = response
    
    
    
    ###### PARSING

    # show simple statistics of contributions
    response = show_contributions(logger, dbuni, args.goodbefore)
    if response == 1: return 1
    
    
    # create the model
    universe = cobra.Model('universe')   
    
    
    # introduce M 
    universe = introduce_metabolites(logger, dbuni, universe, idcollection_dict, kegg_compound_to_others, args.outdir, args.goodbefore[0], args.onlycurator)
    if type(universe)==int: return 1

    
    # introduce R
    universe = introduce_reactions(logger, dbuni, universe, idcollection_dict, kegg_reaction_to_others, args.outdir, args.goodbefore[1], args.onlycurator)
    if type(universe)==int: return 1


    # introduce T
    universe = introduce_transporters(logger, dbuni, universe, idcollection_dict, kegg_reaction_to_others, args.outdir, args.goodbefore[2], args.onlycurator)
    if type(universe)==int: return 1
    

    # introduce sinks / demands (exchanges where included during T)
    universe = introduce_sinks_demands(logger, universe, dbuni)
    if type(universe)==int: return 1


    # introducce universal biomass
    universe = introduce_universal_biomass(logger, universe, dbuni)
    if type(universe)==int: return 1


    # translate Gs to symbols and annotate them (EC, COG, GO, ...)
    universe = translate_annotate_genes(logger, universe, idcollection_dict)
    if type(universe)==int: return 1


    # introduce collections (groups of Rs as maps/modules)
    universe = set_up_groups(logger, universe, idcollection_dict, args.goodbefore, dbuni)
    if type(universe)==int: return 1


    # introduce hand-drwn escher map annotation
    if args.escherzip != '-':
        universe = parse_zipped_escher(logger, universe, args.escherzip, args.outdir)
        if type(universe) == int: return 1


    
    ###### POST-PARSING
    
    # log metrics
    log_metrics(logger, universe)
    
    
    # check absence of unbalancing above the threshold
    log_unbalances(logger, universe)

    
    # check universe completness
    df_C = check_completeness(logger, universe, args.progress, args.module, args.focus, args.taxon, args.eggnog, args.keggorg, idcollection_dict, summary_dict, args.outdir)
    if type(df_C)==int: return 1

    
    # avoid time-consuming activities
    if not args.justparse:
    
    
        # check erroneous EGCs
        verify_egc_all(logger, universe, args.outdir)

        
        # check growth on minmal media
        df_G = grow_on_media(logger, universe, dbexp, args.media, fva=False, universe_in_parsedb=True)
        if type(df_G)==int: return 1

    
        # check blocked biomass precursors
        cond_col_dict = adjust_biomass_precursors(logger, universe, universe, 1.0, dbuni)
        df_E = precursors_on_media(logger, universe, universe, dbuni, dbexp, args.media, cond_col_dict, args.precursors)
        if type(df_E)==int: return 1

    
        # check blocked metabolites / dead-ends
        df_S = biosynthesis_on_media(logger, universe, dbexp, args.media, args.biosynth)
        if type(df_S)==int: return 1


        # reset growth environment befor saving the model (changed during growth sims)
        reset_growth_env(universe)

        
        # initialize model
        response = initialize_model(logger, universe, dbexp, args.initialize, args.media)
        if response==1: return 1


        # compute Memote metrics
        memote_results_dict = get_memote_results_dict(logger, universe)


    # write JSON 
    logger.info("Writing universal model...")
    cobra.io.save_json_model(universe, f'{args.outdir}/universe.json')
    logger.info(f"'{args.outdir}/universe.json' created!")
    
    
    # avoid time-consuming activities
    if not args.justparse:
        
        
        # write XML
        cobra.io.write_sbml_model(universe, f'{args.outdir}/universe.xml')   # groups are saved only to SBML 
        logger.info(f"'{args.outdir}/universe.xml' created!")
        force_id_on_sbml(f'{args.outdir}/universe.xml', 'universe')   # force introduction of the 'id=""' field
        
        
        # write XLSX
        sheets_dict = write_excel_model(universe, f'{args.outdir}/universe.parsedb.xlsx', args.nofigs, memote_results_dict, df_E, None, None, df_S, df_C)  
        logger.info(f"'{args.outdir}/universe.parsedb.xlsx' created!")

    
    # check if escher map is updated:
    if args.focus != '-':
        lastmap = get_eschermap(logger, dbuni, args.focus)
        count_undrawn_rids_focus(logger, universe, lastmap, args.focus, args.outdir)
    
    
    return 0


import warnings
import logging

from cobra.util.solver import linear_reaction_coefficients



def get_objectives(model):
    
    
    objs = list(linear_reaction_coefficients(model).keys())
    obj_ids = [obj.id for obj in objs]
    return obj_ids



def get_optthr():
    
    return 0.001   # optimization threshold



def fba_no_warnings(model): 
    
    # Ignore eventual "UserWarning: Solver status is 'infeasible'."
    with warnings.catch_warnings():  # temporarily suppress warnings for this block
        warnings.simplefilter("ignore")  # ignore all warnings
        
        # disable warnings
        cobra_logger = logging.getLogger("cobra.util.solver")
        old_level = cobra_logger.level
        cobra_logger.setLevel(logging.ERROR)   

        # perform FBA: 
        res = model.optimize()
        obj_value = res.objective_value
        status = res.status

        # restore original behaviour: 
        cobra_logger.setLevel(old_level)   

        return res, obj_value, status
    
    
    
def verify_growth(model, boolean=True):
            
    
    res, obj_value, status = fba_no_warnings(model)
    if boolean:
        if obj_value < get_optthr() or status=='infeasible':
            return False
        else: return True
    else:
        if status =='infeasible':
            return 'infeasible'
        elif obj_value < get_optthr():
            return 0.0
        else:
            return round(obj_value, 3)
        
        

def sensitivity_analysis(model, scaled=False, top=3, mid=None):
    """Perform a sensitivity analysis (or reduced costs analysis) focused on the EX_change reaction.
    
    It is based on the current model's objective. The returned dictionary is sorted from most negative to most positive values.
    
    Args:
        model (cobra.Model): target model.
        scaled (bool): whether to scale to the current objective value.
        top (int): get just the first and last `top` EX_change reactions. If ``None``, all EX_change reactions will be returned.
        mid (str): instead of optimizing for the current objective reaction, do the analysis on the biosynthesis of a specific metabolite having ID `mid`. If `None` it will be ignored.
    
    Returns:
        dict: reduced costs keyd by EX_change reaction ID. 
    """
    
    
    # temporary chenges:
    with model:
        
      
        # focus on a specific metbaolite
        if mid != None: 
            model.objective = add_demand(model, mid)


        res = model.optimize()
        obj = res.objective_value
        flx = res.fluxes.to_dict()
        rcs = res.reduced_costs.to_dict()


        # manage 0 flux exception:
        if obj == 0 and scaled == True:
            raise Exception("Cannot scale reduced costs id the objective value is 0")


        # get the reduced costs of the EXR only:
        rcsex = {} 
        for key in rcs:
            if key.startswith("EX_"):
                if not scaled : rcsex[key] = rcs[key]
                else: rcsex[key] = rcs[key] * flx[key] / obj


        # get the most impactful (lowest and highest)
        rcsex = sorted(rcsex.items(), key=lambda x: x[1], reverse=True)
        rcsex = {i[0]: i[1] for i in rcsex}  # convert list to dictionary
    
        
        # get only the top N and bottom N exchanges
        if top != None: 
            rcsex_filt = {}
            for i in range(top):
                rcsex_filt[ list(rcsex.keys())[i] ] = list(rcsex.values())[i]
            for i in range(top):
                rcsex_filt[ list(rcsex.keys())[-top +i] ] = list(rcsex.values())[-top +i]
            rcsex = rcsex_filt
        
        return rcsex
        
        
        












def get_deprecated_kos():
    deprecated_kos = [
        'K11189',  # should be K02784
        'K07011',  # linked to lp_1215(cps3A) and lp_1216(cps3B) during 2018 and not replaced
        #'K24301',   # to be introduced in GPRs
    ]
    return deprecated_kos



def get_krs_to_exclude():
    return set([
        'R12328', 'R05190',  # general forms of fatty acid biosynthesis
        'R01347', 'R01348', 'R04121',  # general forms of fatty acid degradation
        'R11671',  # multi-step fatty acids reactions
        'R07860', 'R01317', 'R07064',  # aspecific fatty acid reactions
        'R11311', 'R11256', 'R11308', 'R08772', 'R08770',  # polymer reactions
        
        # inconclusive due to semplification
        'R12425', 
        
        # "incomplete reaction" / "unclear reaction"
        'R08414', 'R13037', 'R13034', 'R13036', 'R02825', 'R11178', 'R13325', 'R12855', 'R12856', 'R09809', 
        'R09808', 'R08035', 'R08034', 'R11470', 'R09360', 'R08139', 'R08318', 'R07859', 'R09361', 'R09349', 
        'R13149', 'R13066', 'R11467', 'R11255', 'R08986', 'R13156', 'R13074', 'R13150', 'R11302', 'R11388', 
        'R08341', 'R13147', 'R13155', 'R08339', 'R11466', 'R08272', 'R09348', 'R09362', 'R11107', 'R08340', 
        'R07940', 'R11120', 'R11245', 'R08269', 'R11131', 'R07943', 'R08342', 'R06766', 'R12584', 'R09852',
        'R08268', 'R11129', 'R06702', 'R08866', 'R12555', 'R08927', 'R08343', 'R13067', 'R13069', 'R13068',
        'R05670', 'R06694', 'R09851', 'R11465', 'R08928', 'R11389', 'R11464', 'R13087', 'R12586', 'R11304', 
        'R08984', 'R11254', 'R13165', 'R12884', 'R08865', 'R13151', 'R08132', 'R08929', 'R06701', 'R08345',
        'R11365', 'R11303', 'R06670', 'R11364', 'R09347', 'R08293', 'R11362', 'R03872', 'R06339', 'R10481', 
        'R10480', 'R13341', 'R06505', 'R06504', 'R06326', 'R06470', 'R06467', 'R06327', 'R06503', 'R09847',
        'R13479', 'R13447', 'R13478', 'R07510', 'R04546', 'R06468', 'R05624', 'R10706', 'R13454', 'R13556',
        'R13455', 'R12691', 
    ])



def get_rids_with_mancheck_gpr():
    rids_mancheck_gpr = [  # reactions with manually checked GPRs
        'SUCD1', 'ALKP', 'PFK_3', 'TCMPTS', 'PPA', 'APSR',
        'NADHPO', 'ACOD', '5DOAN', '5DRBK', '5DRBPI', '5DRBUPD',
        'FNPOR', 'THZPSN', 'PPTGP', 'ALATA_D2', 'XYLUR1'
    ]
    return rids_mancheck_gpr



def get_rids_with_mancheck_balancing():
    rids_mancheck_bal = [  # same reactions involving ATP can be reversible
        
        # SECTION "reversible both in KEGG and MetaCyc"
        'PGK', 'SUCOAS', 'ADK1', 'GK1', 'NNATr', 'CYTK1', 'ACKr',
        'DGK1', 'PPAKr', 'ATPSr', 'NDPK10', 'BUTKr',
        
        ### SECTION "reversible in KEGG but not in MetaCyc" ###
        'CYTK2',  # clearly reversible in KEGG but not in MetaCyc (RXN-7913)
        'DADK',  # clearly reversible in KEGG but not in MetaCyc (DEOXYADENYLATE-KINASE-RXN)
        'UMPK',  # clearly reversible in KEGG but not in MetaCyc (RXN-12002)
        'NDPK1',  # clearly reversible in KEGG but not in MetaCyc (GDPKIN-RXN)
        'NDPK2',  # clearly reversible in KEGG but not in MetaCyc (UDPKIN-RXN)  
        'NDPK3',  # clearly reversible in KEGG but not in MetaCyc (CDPKIN-RXN)
        'NDPK4',  # clearly reversible in KEGG but not in MetaCyc (DTDPKIN-RXN)
        'NDPK5',  # clearly reversible in KEGG but not in MetaCyc (DGDPKIN-RXN)
        'NDPK6',  # clearly reversible in KEGG but not in MetaCyc (DUDPKIN-RXN)
        'NDPK7',  # clearly reversible in KEGG but not in MetaCyc (DCDPKIN-RXN)
        'NDPK8',  # clearly reversible in KEGG but not in MetaCyc (DADPKIN-RXN)
        'NDPK9',  # clearly reversible in KEGG but not in MetaCyc (RXN-14120) 
        
        ### SECTION "missing reversibility info" ###
        'LPHERA',  
    ]
    return rids_mancheck_bal





import warnings
import multiprocessing

import cobra
from cobra.flux_analysis.gapfilling import GapFiller

from ..commons.fluxbal import fba_no_warnings
from ..commons.fluxbal import get_optthr



def add_demand(model, mid):
    
    
    rid = f"demand_{mid}"
    newr = cobra.Reaction(rid)
    model.add_reactions([newr])
    model.reactions.get_by_id(rid).reaction = f"{mid} -->"
    model.reactions.get_by_id(rid).bounds = (0, 1000)
    
    return rid



def can_synth(model, mid):
    
    
    # changes are temporary: demand is not added, objective is not changed.
    with model: 

        rid = add_demand(model, mid)

        # set the objective to this demand reaction:
        model.objective = rid

        # perform FBA: 
        res, value, status = fba_no_warnings(model)
        response = True if (value > get_optthr() and status == 'optimal') else False
        
        return response, round(value, 2), status



def perform_gapfilling(model, universe, mid=None, slim=None, minflux=1, exr=False, nsol=3, penalties=None, verbose=True, timeout=None, logger=None, boost=False): 

    
    def temporized_task(model, universe, minflux, exr, nsol, penalties):

        # launch cobrapy gapfilling: 
        gapfiller = GapFiller(model, universe, 
            lower_bound = minflux,
            demand_reactions = False, 
            exchange_reactions = exr, 
            integer_threshold = 1e-10,  # default: 1e-6.
            penalties = penalties,
        )
        try: solutions = gapfiller.fill(iterations=nsol)
        except Exception as e: 
            return e
        
        return solutions
            
       
    # temporary changes (objective, solver, demands edits are temporary)
    # remove genes to avoid errors due to the context restoration (eg "ValueError: id {gene_id} is already present in list")
    model_nogenes = model.copy()
    cobra.manipulation.remove_genes(model_nogenes, [g for g in model_nogenes.genes], remove_reactions=False)
    universe_nogenes = universe.copy()
    cobra.manipulation.remove_genes(universe_nogenes, [g for g in universe_nogenes.genes], remove_reactions=False)
    

    # if focusing on a particular biosynthesis
    if mid != None:
        model_nogenes.objective = add_demand(model_nogenes, mid)
        universe_nogenes.objective = add_demand(universe_nogenes, mid)


    # if requested, boost the actual sources:
    if boost:
        for r in model_nogenes.reactions: 
            if len(r.metabolites)==1 and list(r.metabolites)[0].id.endswith('_e'): # if exchange reaction
                if r.lower_bound < 0: 
                    r.lower_bound = -1000


    # "AssertionError: daemonic processes are not allowed to have children" could be raised since
    # child processes created by multiprocessing.Pool are demoniac by deafault,
    # and demoniac processes cannot create their own child processes like the temporized gap-filler. 
    if timeout != None: 


        # run the gapfilling algorithm inside a timer: 
        # first define the worker for the separate process
        def worker(results_channel):
            result = temporized_task(model_nogenes, universe_nogenes, minflux, exr, nsol, penalties)
            results_channel.put(result)


        # then start the separate process
        results_channel = multiprocessing.Queue()
        process = multiprocessing.Process(target=worker, args=(results_channel,))
        process.start()  # start the process
        process.join(timeout) # wait for the process to finish or timeout


        # if still running after the timeout:
        if process.is_alive():  
            process.terminate()
            e = "Gap-filling timed out, consider to increase the 'timeout' attribute."
            if verbose:  # avoid the error stack trace
                print("WARNING: cobrapy:", e) if logger==None else logger.warning(f'cobrapy: {e}')  
            return None


        else: # the worker finished: retrieve the result or exception from the queue:
            result = results_channel.get()
            if isinstance(result, Exception):  
                if verbose: 
                    print("WARNING: cobrapy:", result) if logger==None else logger.warning(f'cobrapy: {result}')  
                return None  
            else:  # yielded a true result, not an Exception
                solutions = result


    else:  # do not temporize the gap-filling:
        result = temporized_task(model_nogenes, universe_nogenes, minflux, exr, nsol, penalties)
        if isinstance(result, Exception):  
            if verbose: 
                print("WARNING: cobrapy:", result) if logger==None else logger.warning(f'cobrapy: {result}')
            return None
        else:  # yielded a true result, not an Exception
            solutions = result


    # iterate the solutions:
    first_sol_rids = []  # rids proposed during the 1st solution
    for i, solution in enumerate(solutions):
        if verbose and logger==None: print(f'Solution {i+1}. Reactions to add: {len(solution)}.')


        # iterate the reactions: 
        counter = 0
        for r in solution: 
            counter += 1
            # note: this 'r' is not linked to any model. 
            # indeed, calling r.model, None will be returned. 
            if verbose and logger==None: print(f'{counter} {r.id} {r.name}')

            
            # populate results with IDs from first solution:
            if i == 0: first_sol_rids.append(r.id)

            
        # separate solutions with a new line:
        if i+1 != len(solutions): 
            if verbose and logger==None: print()


    return first_sol_rids



def import_from_universe(model, universe, rid, bounds=None, gpr=None):

    
    # get the universal reaction
    ru = universe.reactions.get_by_id(rid)
    
    # create a new empty reaction
    r = cobra.Reaction(rid)
    model.add_reactions([r])
    r = model.reactions.get_by_id(rid)
    
    # copy the name
    r.name = ru.name
    
    # build string all universal metabolites are still there
    # (remove__disconnected is called later):
    r.build_reaction_from_string(ru.reaction)
    
    # set bounds
    if bounds != None: r.bounds = bounds
    else: r.bounds = ru.bounds
        
    # set GPR
    if gpr != None:
        r.gene_reaction_rule = gpr
    else:
        r.gene_reaction_rule = ''
    r.update_genes_from_gpr()
    
    # set annotations
    r.annotation = ru.annotation



def include_forced(logger, model, universe, force_inclusion):
    
    
    # get rids provided by user:
    if force_inclusion != '-':
        introduced_rids = []
        forced_rids = force_inclusion.split(',')
        
        
        # get rids already modeled
        modeled_rids = [r.id for r in model.reactions]
        universal_rids = [r.id for r in universe.reactions]
        
        
        # skip reactions not modeled in whole universe
        for rid in forced_rids: 
            if rid not in universal_rids:
                logger.debug(f"Ignoring reaction '{rid}' since it's not included in the universe.")
                continue
            
            
            # otherwise add the universal reaction
            if rid not in modeled_rids:
                import_from_universe(model, universe, rid, gpr='')
                introduced_rids.append(rid)
                logger.warning(f"Reaction '{rid}' forcibly included as orphan.")
            else:
                logger.debug(f"Requested reaction '{rid}' was already included.")
        logger.info(f"Forcibly included {len(introduced_rids)} reactions as orphans.")
        
        

def get_repository_nogenes(logger, universe):
    # provide a gene-free repository of reactions to be used during gap-filling


    # make an editable copy:
    repository_nogenes = universe.copy()
    
        
    # remove all genes to avoid the "ValueError: id purP is already present in list"
    cobra.manipulation.delete.remove_genes(repository_nogenes, [g.id for g in repository_nogenes.genes], remove_reactions=False)

    
    return repository_nogenes




    
    




import threading
import copy

import cobra

from .manual import get_deprecated_kos

from ..commons.downloads import SimpleLoadingWheel



def translate_annotate_genes(logger, model, idcollection_dict):
    
    
    logger.info("Translating and annotating orthologs...")
    
    
    # extract idcollections
    ko_to_name = idcollection_dict['ko_to_name']
    ko_to_symbols = idcollection_dict['ko_to_symbols']
    ko_to_ecs = idcollection_dict['ko_to_ecs']
    ko_to_cogs = idcollection_dict['ko_to_cogs']
    ko_to_gos = idcollection_dict['ko_to_gos']
    ko_to_taxa = idcollection_dict['ko_to_taxa']
    
    
    # create the translation dicts (ko_to_sym): 
    # assign to each KO a symbol that is unique in the universe model.
    ko_to_sym = {}
    sym_to_ko = {}
    cnt = 0
    for g in model.genes:
        if g.id in ['orphan', 'spontaneous']: 
            continue
        ko = g.id
        cnt += 1
        
        
        # if the ko is deprecated, it was not included in 'ko_to_symbols'
        if ko in get_deprecated_kos():
            ko_to_sym[ko] = ko
            sym_to_ko[ko] = ko
            continue
        
        
        # iterate the available symbols for this KO
        for symbol in ko_to_symbols[ko]:  
            # take the first available (not yet used)
            if symbol not in sym_to_ko.keys():   
                ko_to_sym[ko] = symbol
                sym_to_ko[symbol] = ko
                break
        
        
        # no symbol was assigned (symbol was already taken by another KO)
        if cnt != len(ko_to_sym):  
            cnt_dups = 2
            symbol = list(ko_to_symbols[ko])[0] + f'_{cnt_dups}'   # generate a new symbol
            while cnt != len(ko_to_sym):   # until a symbol is assigned
                if symbol not in sym_to_ko.keys():   # if the new symbol fits
                    ko_to_sym[ko] = symbol
                    sym_to_ko[symbol] = ko
                cnt_dups += 1
                symbol = list(ko_to_symbols[ko])[0] + f'_{cnt_dups}'   # retry with the next one
                

    
    # insert annotations
    for g in model.genes:
        if g.id in ['orphan', 'spontaneous']: 
            continue
        ko = g.id
        g.annotation['ko'] = ko
        
        
        if ko not in get_deprecated_kos():
            # deprecated kos are missing from these dicts
            g.name = ko_to_name[ko]
            g.annotation['symbols'] = list(ko_to_symbols[ko])
            g.annotation['ec'] = list(ko_to_ecs[ko])
            g.annotation['cog'] = list(ko_to_cogs[ko])
            g.annotation['go'] = list(ko_to_gos[ko])
            
            
            # add taxa information
            g.annotation['kingdom'] = list(ko_to_taxa[ko]['kingdom'])
            g.annotation['phylum'] = list(ko_to_taxa[ko]['phylum'])
                
            
        # add SBO annotation
        g.annotation['sbo'] = ['SBO:0000243']  # demand reaction 
        
    
        
    # handle orphan and spontaneous
    translation_dict = ko_to_sym
    translation_dict['orphan'] = 'orphan'
    translation_dict['spontaneous'] = 'spontaneous'
    
    
    # finally apply translations of IDs in a dedicated Thread
    t1 = threading.Thread(target = cobra.manipulation.rename_genes, args=(
        model, translation_dict))
    t1.start()
    slw = SimpleLoadingWheel(msg="Please wait... ")
    while t1.is_alive():
        slw.proceed()
    slw.clear()
    
    
    return model
    


def set_up_groups(logger, model, idcollection_dict, goodbefore, dbuni):
    
    
    logger.debug("Introducing groups...")
    modeled_rids = [r.id for r in model.reactions]
    
    
    # extract dicollections
    kr_to_maps = idcollection_dict['kr_to_maps']
    map_to_name = idcollection_dict['map_to_name']
    kr_to_mds = idcollection_dict['kr_to_mds']
    md_to_name = idcollection_dict['md_to_name']
    
    
    # define groups of available contents (Maps + Modules)
    groups = {}   # mixing maps and mds
    for r in model.reactions:
        if 'kegg.reaction' not in r.annotation.keys():
            continue   # Biomass, exchanges, demands, sinks, transporters
        kr_ids = r.annotation['kegg.reaction']
        for kr_id in kr_ids:
            if kr_id == 'RXXXXX':
                continue
            # insert maps
            for map_id in kr_to_maps[kr_id]:
                if map_id not in groups.keys():
                    groups[map_id] = set()
                groups[map_id].add(r)
            # insert mds
            for md_id in kr_to_mds[kr_id]:
                if md_id not in groups.keys():
                    groups[md_id] = set()
                groups[md_id].add(r)
                
                
    # finally insert groups
    for group_id in groups.keys():
        # get group name
        if group_id.startswith('map'):
            name = map_to_name[group_id]
        if group_id.startswith('M'):
            name = md_to_name[group_id]
        actual_group = cobra.core.Group(
            group_id, 
            name = name,
            members = list(groups[group_id]),
            kind = 'partonomy',
        )
        model.add_groups([actual_group])
        
        
    # create a group for transporters on-the-fly.
    # read custom groups from the database: 
    custom_groups = {}
    G = dbuni['G'].set_index('gr_id', drop=True, verify_integrity=True)
    for gr_id, row in G.iterrows():
        rids = [i.strip() for i in row['rids'].strip().rsplit(';')]
        custom_groups[gr_id] = rids
    
    
    # add 'gr_transport'
    custom_groups['gr_transport'] = []
    for r in model.reactions:
        if len(r.metabolites) == 1 and list(r.metabolites)[0].id.rsplit('_',1)[-1] != 'c':  # just exchanges (esclude sinks/demands)
            custom_groups['gr_transport'].append(r.id)
        if len(set([m.id.rsplit('_', 1)[-1] for m in r.metabolites])) > 1:  # transport reactions
            custom_groups['gr_transport'].append(r.id)
    
    
    # handle eventual removed reactions as consequence of --goodbefore:
    if type(goodbefore) == list:
        for group_id in custom_groups.keys():
            for rid in copy.deepcopy(custom_groups[group_id]):
                if rid not in modeled_rids:
                    logger.warning(f"Removing reaction '{rid}' from group '{group_id}' as consequence of --goodbefore.")
                    custom_groups[group_id].remove(rid)
        
        
    # actually insert custom groups:
    for group_id in custom_groups.keys():
        actual_group = cobra.core.Group(
            group_id, 
            name = group_id,
            members = [model.reactions.get_by_id(rid) for rid in custom_groups[group_id]],
            kind = 'partonomy',
        )
        model.add_groups([actual_group])
        
        
    return model



import requests
import io
import threading
import time
import tempfile
import json
import glob
import sys

import gdown
import pandas as pnd



class SimpleLoadingWheel():
    # a simple endless loading wheel
    
    msg = ''
    wheel_figures = ['|', '/', '-', '\\']
    wheel_cnt = 0
    
    def __init__(self, msg='Please wait... '):
        self.msg = msg
    
    def proceed(self):
        print(f"{self.msg}{self.wheel_figures[self.wheel_cnt]}", file=sys.stderr, end='\r')
        self.wheel_cnt += 1
        if self.wheel_cnt == 4: self.wheel_cnt = 0
        time.sleep(0.5)
        
    def clear(self):
        clear_str = ''.join([' ' for i in range(len(self.msg)+1)])
        print(f'{clear_str}', file=sys.stderr, end='\r')
        wheel_cnt = 0
    


def get_dbuni(logger):
    
    
    sheet_id = "1dXJBIFjCghrdvQtxEOYlVNWAQU4mK-nqLWyDQeUZqek"
    #sheet_id = "15fIBewG1B1jIbg1_9pMnyPJL7LWCLMD8s_vlf3ZUno0"  # alternative
    url = f"https://docs.google.com/spreadsheets/d/{sheet_id}/export?format=xlsx"
    response = requests.get(url)  # download the requested file
    if response.status_code == 200:
        excel_data = io.BytesIO(response.content)   # load into memory
        exceldb = pnd.ExcelFile(excel_data)
    else:
        logger.error(f"Error during download. Retry! If persists, please contact the developer.")
        return 1
    
    
    # check table presence
    sheet_names = exceldb.sheet_names
    for i in ['M', 'R', 'T', 'A', 'G', 'B', 'curators']: 
        if i not in sheet_names:
            logger.error(f"Sheet '{i}' is missing!")
            return 1
        
        
    # load the tables
    dbuni = {}
    dbuni['M'] = exceldb.parse('M')
    dbuni['R'] = exceldb.parse('R')
    dbuni['T'] = exceldb.parse('T')
    dbuni['A'] = exceldb.parse('A')
    dbuni['G'] = exceldb.parse('G')
    dbuni['B'] = exceldb.parse('B')
    dbuni['E'] = exceldb.parse('E')
    dbuni['curators'] = exceldb.parse('curators')
    
    
    # check table headers
    headers = {}
    headers['M'] = ['pure_mid', 'formula', 'charge', 'kc', 'name', 'inchikey', 'curator', 'notes']
    headers['R'] = ['rid', 'rstring', 'kr', 'gpr_manual', 'name', 'curator', 'notes']
    headers['T'] = ['rid', 'rstring', 'kr', 'gpr_manual', 'name', 'curator', 'notes']
    headers['A'] = ['pure_mid', 'type', 'curator', 'notes']
    headers['G'] = ['gr_id', 'rids', 'curator', 'notes']
    headers['B'] = ['pure_mid', 'class', 'evaluation', 'bs_groups', 'curator', 'notes']
    headers['E'] = ['map_id', 'location', 'curator', 'notes']
    headers['curators'] = ['username', 'first_name', 'last_name', 'role', 'mail']
    for i in dbuni.keys(): 
        if set(dbuni[i].columns) != set(headers[i]):
            logger.error(f"Sheet '{i}' is missing the columns {set(headers[i]) - set(dbuni[i].columns)}.")
            return 1
        
    return dbuni



def get_dbexp(logger):
    
    
    sheet_id = "1qGbIIipHJgYQjk3M0xDWKvnTkeInPoTeH9unDQkZPwg"
    #sheet_id = "1qxTRf30SeT9WJFYxWm2ChCxkTR0sTn7BbDOFhUuUQIE"   # alternative
    url = f"https://docs.google.com/spreadsheets/d/{sheet_id}/export?format=xlsx"
    response = requests.get(url)  # download the requested file
    if response.status_code == 200:
        excel_data = io.BytesIO(response.content)   # load into memory
        exceldb = pnd.ExcelFile(excel_data)
    else:
        logger.error(f"Error during download. Retry! If persists, please contact the developer.")
        return 1
    
    
    # check table presence
    sheet_names = exceldb.sheet_names
    for i in ['media', 'PM1', 'PM2A', 'PM3B', 'PM4A', 'MWF', 'DNA', 'RNA', 'PROTS', 'LIPIDS_PL', 'LIPIDS_FA', 'curators']: 
        if i not in sheet_names:
            logger.error(f"Sheet '{i}' is missing!")
            return 1
        
        
    # load the tables
    dbexp = {}
    dbexp['media'] = exceldb.parse('media')
    dbexp['PM1'] = exceldb.parse('PM1')
    dbexp['PM2A'] = exceldb.parse('PM2A')
    dbexp['PM3B'] = exceldb.parse('PM3B')
    dbexp['PM4A'] = exceldb.parse('PM4A')
    dbexp['MWF'] = exceldb.parse('MWF')
    dbexp['DNA'] = exceldb.parse('DNA')
    dbexp['RNA'] = exceldb.parse('RNA')
    dbexp['PROTS'] = exceldb.parse('PROTS')
    dbexp['LIPIDS_PL'] = exceldb.parse('LIPIDS_PL')
    dbexp['LIPIDS_FA'] = exceldb.parse('LIPIDS_FA')
    dbexp['curators'] = exceldb.parse('curators')
    
    
    # format tables (media):
    # assign substrates as index
    dbexp['media'].index = dbexp['media'].iloc[:, 1]
    # remove first 2 useless column (empty & substrates)
    dbexp['media'] = dbexp['media'].iloc[:, 2:]
    
    
    # format tables (Biolog(R)):
    for sheet in ['PM1', 'PM2A', 'PM3B', 'PM4A']:
        # assign wells as index
        dbexp[sheet].index = dbexp[sheet].iloc[:, 2]
        # remove first 3 useless columns
        dbexp[sheet] = dbexp[sheet].iloc[:, 3:]
        
        
    # format tables (biomass):
    dbexp['MWF'].index = dbexp['MWF'].iloc[:, 0]   # assign index
    dbexp['MWF'] = dbexp['MWF'].iloc[:, 1:]        # remove meaningless columns
    #
    dbexp['DNA'].index = dbexp['DNA'].iloc[:, 0]   # assign index
    dbexp['DNA'] = dbexp['DNA'].iloc[:, 1:]        # remove meaningless columns
    #
    dbexp['RNA'].index = dbexp['RNA'].iloc[:, 0]   # assign index
    dbexp['RNA'] = dbexp['RNA'].iloc[:, 1:]        # remove meaningless columns
    # 
    dbexp['PROTS'].index = dbexp['PROTS'].iloc[:, 1]   # assign index
    dbexp['PROTS'] = dbexp['PROTS'].iloc[:, 2:]        # remove meaningless columns
    #
    dbexp['LIPIDS_PL'].index = dbexp['LIPIDS_PL'].iloc[:, 1]   # assign index
    dbexp['LIPIDS_PL'] = dbexp['LIPIDS_PL'].iloc[:, 2:]        # remove meaningless columns
    #
    dbexp['LIPIDS_FA'].index = dbexp['LIPIDS_FA'].iloc[:, 1]   # assign index
    dbexp['LIPIDS_FA'] = dbexp['LIPIDS_FA'].iloc[:, 2:]        # remove meaningless columns
    
    
    return dbexp



def get_eschermap(logger, dbuni, map_id):
    
    
    # retrieve catalog from database:
    dict_escher_maps = {}
    E = dbuni['E'].set_index('map_id', drop=True, verify_integrity=True)
    for index, row in E.iterrows():
        dict_escher_maps[index] = row['location']
    
    
    # return None if not requested
    if map_id == None or map_id == '-': 
        return None
    
    
    # check the esistance of the map  
    if map_id not in dict_escher_maps.keys():
        logger.warning(f"No online folder is tracked for '{map_id}' yet. Please edit our database (sheet 'E') to track '{map_id}'.")
        return None    
    
    
    # object to be returned
    logger.info(f"Downloading last version of map '{map_id}'...")
    lastmap = dict()  
    
    
    # the temporary folder (/tmp/on-the-fly-code) will be deleted once exiting the with statement
    with tempfile.TemporaryDirectory() as tmp_dir:
        
        
        # get available versions without downloading:
        folder_id = dict_escher_maps[map_id]
        folder_url = f"https://drive.google.com/drive/folders/{folder_id}?usp=sharing"
        contents = gdown.download_folder(folder_url, output=tmp_dir, quiet=True, skip_download=True)

        # check empty folder
        if len(contents) == 0:
            logger.warning(f"Online folder for '{map_id}' seems empty. Please draw and upload '{map_id}-v1.json'.")
            return None
        
        # check crowded folder
        if len(contents) > 40:
            logger.warning(f"Online folder for '{map_id}' contains >40 versions. An error will be raised at 50 versions. Please free space by moving older versions into 'older/' now.")
        
        # create dict of available files:
        files = {i.path: i.id for i in contents}
        
        # check file name consistency
        map_ids = set([i.rsplit('-v',1)[0] for i in files.keys()])
        if len(map_ids) != 1:
            logger.error(f"Several different map names in online folder for '{map_id}'.")
            return 1
        
        # check concordance to mother folder
        if list(map_ids)[0] != map_id:
            logger.error(f"Online folder for '{map_id}' contains versions for another map.")
            return 1
        
        # get the last available version (natural order)
        last_version = max([int(i.rsplit('-v',1)[-1].replace('.json', '')) for i in files.keys()])
        last_file_name = list(files.keys())[0].rsplit('-v',1)[0] + f"-v{last_version}.json"
        last_file_id = files[last_file_name]
        lastmap['filename'] = last_file_name
        
        # download last version:
        try: pathfile_tmpfolder = gdown.download(id=last_file_id, output=f"{tmp_dir}/{last_file_name}", quiet=True)
        except: 
            logger.error(f"Downloading of last-version for '{map_id}'. Retry. If persists, please contact the developer.")
            return 1
        
        # load json
        with open(pathfile_tmpfolder, 'r') as file:
            json_data = json.load(file)
            lastmap['json'] = json_data

        
    return lastmap
        
    
    
def get_databases(logger): 
    

    # define a function to fill the dict
    def run_with_result(func, logger, results_dict, key):
        result = func(logger)
        results_dict[key] = result
        
    
    # start threads
    results_dict = dict()
    t1 = threading.Thread(target=run_with_result, args=(
        get_dbuni, logger, results_dict, 'dbuni'))
    t2 = threading.Thread(target=run_with_result, args=(
        get_dbexp, logger, results_dict, 'dbexp'))

    
    # wait for the longest download:
    t1.start()
    t2.start()
    slw = SimpleLoadingWheel(msg="Please wait... ")
    while t1.is_alive() or t2.is_alive():
        slw.proceed()
    slw.clear()
    
    
    # check if errors where raised:
    if type(results_dict['dbuni'])==int:
        return 1
    if type(results_dict['dbexp'])==int:
        return 1
    
    
    return (results_dict['dbuni'], results_dict['dbexp'])



def eggnog_format_expansion(logger, eggnog):
    
    
    # linux/macos usually perform argument axpansion befor passing the parameter to python.
    if type(eggnog) == list: # already expanded by the terminal
        if len(eggnog)==1 and '*' in eggnog[0]:
            original_eggnog = eggnog[0]
            eggnog = glob.glob(eggnog[0])  # glob will append only existing files to the list
            if eggnog == []:
                logger.info(f"No file matching '{original_eggnog}'.")
    
    
    elif type(eggnog) == str:  # in the terminal, it can be specified by using single quotes
        original_eggnog = False
        if eggnog != '-': # user wanted to specify something
            original_eggnog = eggnog
        eggnog = glob.glob(eggnog)  # glob will append only existing files to the list
        if original_eggnog and eggnog == []:
            logger.info(f"No file matching '{original_eggnog}'.")
            
            
    if eggnog == [] or eggnog == ['-']:
        eggnog = '-'   # return always a list except for this case
        
        
    return eggnog
    


def check_taxon(logger, taxon, idcollection_dict):
    
    
    # verify presence of needed assets
    if 'ko_to_taxa' not in idcollection_dict.keys():
        logger.error(f"Asset 'ko_to_taxa' not found in 'gsrap.maps'. Please update 'gsrap.maps' with 'gsrap getmaps'.")
        return 1
    
    
    # extract level and name
    try: level, name = taxon.split(':')
    except: 
        logger.error(f"Provided --taxon is not well formatted: '{taxon}'.")
        return 1
    
    
    # compute available levels and check
    avail_levels = set(['kingdom', 'phylum'])            
    if level not in avail_levels:
        logger.error(f"Provided level is not acceptable: '{level}' (see --taxon). Acceptable levels are {avail_levels}.")
        return 1
    
    
    # compute available taxa at input level
    avail_taxa_at_level = set()
    ko_to_taxa = idcollection_dict['ko_to_taxa']
    for ko in ko_to_taxa.keys():
        for taxon_name in ko_to_taxa[ko][level]:
            avail_taxa_at_level.add(taxon_name)
    if name not in avail_taxa_at_level:
        logger.error(f"Provided taxon name is not acceptable: '{name}' (see --taxon). Acceptable taxon names for level '{level}' are {avail_taxa_at_level}.")
        return 1
    
    
    return 0
